# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility function for building Importer Node spec."""

from typing import Union, Type

from kfp.dsl import _container_op
from kfp.dsl import _pipeline_param
from kfp.dsl import dsl_utils
from kfp.pipeline_spec import pipeline_spec_pb2
from kfp.v2.components.types import artifact_types, type_utils

INPUT_KEY = 'uri'
OUTPUT_KEY = 'artifact'


def _build_importer_spec(
    artifact_uri: Union[_pipeline_param.PipelineParam, str],
    artifact_type_schema: pipeline_spec_pb2.ArtifactTypeSchema,
) -> pipeline_spec_pb2.PipelineDeploymentConfig.ImporterSpec:
    """Builds an importer executor spec.

    Args:
      artifact_uri: The artifact uri to import from.
      artifact_type_schema: The user specified artifact type schema of the
        artifact to be imported.

    Returns:
      An importer spec.
    """
    importer_spec = pipeline_spec_pb2.PipelineDeploymentConfig.ImporterSpec()
    importer_spec.type_schema.CopyFrom(artifact_type_schema)

    if isinstance(artifact_uri, _pipeline_param.PipelineParam):
        importer_spec.artifact_uri.runtime_parameter = INPUT_KEY
    elif isinstance(artifact_uri, str):
        importer_spec.artifact_uri.constant_value.string_value = artifact_uri

    return importer_spec


def _build_importer_task_spec(
    importer_base_name: str,
    artifact_uri: Union[_pipeline_param.PipelineParam, str],
) -> pipeline_spec_pb2.PipelineTaskSpec:
    """Builds an importer task spec.

    Args:
      importer_base_name: The base name of the importer node.
      artifact_uri: The artifact uri to import from.

    Returns:
      An importer node task spec.
    """
    result = pipeline_spec_pb2.PipelineTaskSpec()
    result.component_ref.name = dsl_utils.sanitize_component_name(
        importer_base_name)

    if isinstance(artifact_uri, _pipeline_param.PipelineParam):
        param = artifact_uri
        if param.op_name:
            result.inputs.parameters[
                INPUT_KEY].task_output_parameter.producer_task = (
                    dsl_utils.sanitize_task_name(param.op_name))
            result.inputs.parameters[
                INPUT_KEY].task_output_parameter.output_parameter_key = param.name
        else:
            result.inputs.parameters[
                INPUT_KEY].component_input_parameter = param.full_name
    elif isinstance(artifact_uri, str):
        result.inputs.parameters[
            INPUT_KEY].runtime_value.constant_value.string_value = artifact_uri

    return result


def _build_importer_component_spec(
    importer_base_name: str,
    artifact_type_schema: pipeline_spec_pb2.ArtifactTypeSchema,
) -> pipeline_spec_pb2.ComponentSpec:
    """Builds an importer component spec.

    Args:
      importer_base_name: The base name of the importer node.
      artifact_type_schema: The user specified artifact type schema of the
        artifact to be imported.

    Returns:
      An importer node component spec.
    """
    result = pipeline_spec_pb2.ComponentSpec()
    result.executor_label = dsl_utils.sanitize_executor_label(
        importer_base_name)
    result.input_definitions.parameters[
        INPUT_KEY].type = pipeline_spec_pb2.PrimitiveType.STRING
    result.output_definitions.artifacts[OUTPUT_KEY].artifact_type.CopyFrom(
        artifact_type_schema)

    return result


def importer(artifact_uri: Union[_pipeline_param.PipelineParam, str],
             artifact_class: Type[artifact_types.Artifact],
             reimport: bool = False) -> _container_op.ContainerOp:
    """dsl.importer for importing an existing artifact. Only for v2 pipeline.

    Args:
      artifact_uri: The artifact uri to import from.
      artifact_type_schema: The user specified artifact type schema of the
        artifact to be imported.
      reimport: Whether to reimport the artifact. Defaults to False.

    Returns:
      A ContainerOp instance.

    Raises:
      ValueError if the passed in artifact_uri is neither a PipelineParam nor a
        constant string value.
    """

    if isinstance(artifact_uri, _pipeline_param.PipelineParam):
        input_param = artifact_uri
    elif isinstance(artifact_uri, str):
        input_param = _pipeline_param.PipelineParam(
            name='uri', value=artifact_uri, param_type='String')
    else:
        raise ValueError(
            'Importer got unexpected artifact_uri: {} of type: {}.'.format(
                artifact_uri, type(artifact_uri)))

    old_warn_value = _container_op.ContainerOp._DISABLE_REUSABLE_COMPONENT_WARNING
    _container_op.ContainerOp._DISABLE_REUSABLE_COMPONENT_WARNING = True

    task = _container_op.ContainerOp(
        name='importer',
        image='importer_image',  # TODO: need a v1 implementation of importer.
        file_outputs={
            OUTPUT_KEY:
                "{{{{$.outputs.artifacts['{}'].uri}}}}".format(OUTPUT_KEY)
        },
    )
    _container_op.ContainerOp._DISABLE_REUSABLE_COMPONENT_WARNING = old_warn_value

    artifact_type_schema = type_utils.get_artifact_type_schema(artifact_class)
    task.importer_spec = _build_importer_spec(
        artifact_uri=artifact_uri, artifact_type_schema=artifact_type_schema)
    task.task_spec = _build_importer_task_spec(
        importer_base_name=task.name, artifact_uri=artifact_uri)
    task.component_spec = _build_importer_component_spec(
        importer_base_name=task.name, artifact_type_schema=artifact_type_schema)
    task.inputs = [input_param]

    return task



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for input/output Artifacts in KFP SDK.

These are only compatible with v2 Pipelines.
"""

import os
from typing import Dict, Generic, List, Optional, Type, TypeVar, Union

_GCS_LOCAL_MOUNT_PREFIX = '/gcs/'
_MINIO_LOCAL_MOUNT_PREFIX = '/minio/'
_S3_LOCAL_MOUNT_PREFIX = '/s3/'


class Artifact(object):
    """Generic Artifact class.

    This class is meant to represent the metadata around an input or output
    machine-learning Artifact. Artifacts have URIs, which can either be a location
    on disk (or Cloud storage) or some other resource identifier such as
    an API resource name.

    Artifacts carry a `metadata` field, which is a dictionary for storing
    metadata related to this artifact.
    """
    TYPE_NAME = 'system.Artifact'
    VERSION = '0.0.1'

    def __init__(self,
                 name: Optional[str] = None,
                 uri: Optional[str] = None,
                 metadata: Optional[Dict] = None):
        """Initializes the Artifact with the given name, URI and metadata."""
        self.uri = uri or ''
        self.name = name or ''
        self.metadata = metadata or {}

    @property
    def path(self):
        return self._get_path()

    @path.setter
    def path(self, path):
        self._set_path(path)

    def _get_path(self) -> Optional[str]:
        if self.uri.startswith('gs://'):
            return _GCS_LOCAL_MOUNT_PREFIX + self.uri[len('gs://'):]
        elif self.uri.startswith('minio://'):
            return _MINIO_LOCAL_MOUNT_PREFIX + self.uri[len('minio://'):]
        elif self.uri.startswith('s3://'):
            return _S3_LOCAL_MOUNT_PREFIX + self.uri[len('s3://'):]
        return None

    def _set_path(self, path):
        if path.startswith(_GCS_LOCAL_MOUNT_PREFIX):
            path = 'gs://' + path[len(_GCS_LOCAL_MOUNT_PREFIX):]
        elif path.startswith(_MINIO_LOCAL_MOUNT_PREFIX):
            path = 'minio://' + path[len(_MINIO_LOCAL_MOUNT_PREFIX):]
        elif path.startswith(_S3_LOCAL_MOUNT_PREFIX):
            path = 's3://' + path[len(_S3_LOCAL_MOUNT_PREFIX):]
        self.uri = path


class Model(Artifact):
    """An artifact representing an ML Model."""
    TYPE_NAME = 'system.Model'

    def __init__(self,
                 name: Optional[str] = None,
                 uri: Optional[str] = None,
                 metadata: Optional[Dict] = None):
        super().__init__(uri=uri, name=name, metadata=metadata)

    @property
    def framework(self) -> str:
        return self._get_framework()

    def _get_framework(self) -> str:
        return self.metadata.get('framework', '')

    @framework.setter
    def framework(self, framework: str):
        self._set_framework(framework)

    def _set_framework(self, framework: str):
        self.metadata['framework'] = framework


class Dataset(Artifact):
    """An artifact representing an ML Dataset."""
    TYPE_NAME = 'system.Dataset'

    def __init__(self,
                 name: Optional[str] = None,
                 uri: Optional[str] = None,
                 metadata: Optional[Dict] = None):
        super().__init__(uri=uri, name=name, metadata=metadata)


class Metrics(Artifact):
    """Represent a simple base Artifact type to store key-value scalar
    metrics."""
    TYPE_NAME = 'system.Metrics'

    def __init__(self,
                 name: Optional[str] = None,
                 uri: Optional[str] = None,
                 metadata: Optional[Dict] = None):
        super().__init__(uri=uri, name=name, metadata=metadata)

    def log_metric(self, metric: str, value: float):
        """Sets a custom scalar metric.

        Args:
          metric: Metric key
          value: Value of the metric.
        """
        self.metadata[metric] = value


class ClassificationMetrics(Artifact):
    """Represents Artifact class to store Classification Metrics."""
    TYPE_NAME = 'system.ClassificationMetrics'

    def __init__(self,
                 name: Optional[str] = None,
                 uri: Optional[str] = None,
                 metadata: Optional[Dict] = None):
        super().__init__(uri=uri, name=name, metadata=metadata)

    def log_roc_data_point(self, fpr: float, tpr: float, threshold: float):
        """Logs a single data point in the ROC Curve.

        Args:
          fpr: False positive rate value of the data point.
          tpr: True positive rate value of the data point.
          threshold: Threshold value for the data point.
        """

        roc_reading = {
            'confidenceThreshold': threshold,
            'recall': tpr,
            'falsePositiveRate': fpr
        }
        if 'confidenceMetrics' not in self.metadata.keys():
            self.metadata['confidenceMetrics'] = []

        self.metadata['confidenceMetrics'].append(roc_reading)

    def log_roc_curve(self, fpr: List[float], tpr: List[float],
                      threshold: List[float]):
        """Logs an ROC curve.

        The list length of fpr, tpr and threshold must be the same.

        Args:
          fpr: List of false positive rate values.
          tpr: List of true positive rate values.
          threshold: List of threshold values.
        """
        if len(fpr) != len(tpr) or len(fpr) != len(threshold) or len(
                tpr) != len(threshold):
            raise ValueError(
                'Length of fpr, tpr and threshold must be the same. '
                'Got lengths {}, {} and {} respectively.'.format(
                    len(fpr), len(tpr), len(threshold)))

        for i in range(len(fpr)):
            self.log_roc_data_point(
                fpr=fpr[i], tpr=tpr[i], threshold=threshold[i])

    def set_confusion_matrix_categories(self, categories: List[str]):
        """Stores confusion matrix categories.

        Args:
          categories: List of strings specifying the categories.
        """

        self._categories = []
        annotation_specs = []
        for category in categories:
            annotation_spec = {'displayName': category}
            self._categories.append(category)
            annotation_specs.append(annotation_spec)

        self._matrix = []
        for row in range(len(self._categories)):
            self._matrix.append({'row': [0] * len(self._categories)})

        self._confusion_matrix = {}
        self._confusion_matrix['annotationSpecs'] = annotation_specs
        self._confusion_matrix['rows'] = self._matrix
        self.metadata['confusionMatrix'] = self._confusion_matrix

    def log_confusion_matrix_row(self, row_category: str, row: List[float]):
        """Logs a confusion matrix row.

        Args:
          row_category: Category to which the row belongs.
          row: List of integers specifying the values for the row.

         Raises:
          ValueError: If row_category is not in the list of categories
          set in set_categories call.
        """
        if row_category not in self._categories:
            raise ValueError('Invalid category: {} passed. Expected one of: {}'.\
              format(row_category, self._categories))

        if len(row) != len(self._categories):
            raise ValueError('Invalid row. Expected size: {} got: {}'.\
              format(len(self._categories), len(row)))

        self._matrix[self._categories.index(row_category)] = {'row': row}
        self.metadata['confusionMatrix'] = self._confusion_matrix

    def log_confusion_matrix_cell(self, row_category: str, col_category: str,
                                  value: int):
        """Logs a cell in the confusion matrix.

        Args:
          row_category: String representing the name of the row category.
          col_category: String representing the name of the column category.
          value: Int value of the cell.

        Raises:
          ValueError: If row_category or col_category is not in the list of
           categories set in set_categories.
        """
        if row_category not in self._categories:
            raise ValueError('Invalid category: {} passed. Expected one of: {}'.\
              format(row_category, self._categories))

        if col_category not in self._categories:
            raise ValueError('Invalid category: {} passed. Expected one of: {}'.\
              format(row_category, self._categories))

        self._matrix[self._categories.index(row_category)]['row'][
            self._categories.index(col_category)] = value
        self.metadata['confusionMatrix'] = self._confusion_matrix

    def log_confusion_matrix(self, categories: List[str],
                             matrix: List[List[int]]):
        """Logs a confusion matrix.

        Args:
          categories: List of the category names.
          matrix: Complete confusion matrix.

        Raises:
          ValueError: Length of categories does not match number of rows or columns.
        """
        self.set_confusion_matrix_categories(categories)

        if len(matrix) != len(categories):
            raise ValueError('Invalid matrix: {} passed for categories: {}'.\
              format(matrix, categories))

        for index in range(len(categories)):
            if len(matrix[index]) != len(categories):
                raise ValueError('Invalid matrix: {} passed for categories: {}'.\
                  format(matrix, categories))

            self.log_confusion_matrix_row(categories[index], matrix[index])

        self.metadata['confusionMatrix'] = self._confusion_matrix


class SlicedClassificationMetrics(Artifact):
    """Metrics class representing Sliced Classification Metrics.

    Similar to ClassificationMetrics clients using this class are
    expected to use log methods of the class to log metrics with the
    difference being each log method takes a slice to associate the
    ClassificationMetrics.
    """

    TYPE_NAME = 'system.SlicedClassificationMetrics'

    def __init__(self,
                 name: Optional[str] = None,
                 uri: Optional[str] = None,
                 metadata: Optional[Dict] = None):
        super().__init__(uri=uri, name=name, metadata=metadata)

    def _upsert_classification_metrics_for_slice(self, slice: str):
        """Upserts the classification metrics instance for a slice."""
        if slice not in self._sliced_metrics:
            self._sliced_metrics[slice] = ClassificationMetrics()

    def _update_metadata(self, slice: str):
        """Updates metadata to adhere to the metrics schema."""
        self.metadata = {}
        self.metadata['evaluationSlices'] = []
        for slice in self._sliced_metrics.keys():
            slice_metrics = {
                'slice':
                    slice,
                'sliceClassificationMetrics':
                    self._sliced_metrics[slice].metadata
            }
            self.metadata['evaluationSlices'].append(slice_metrics)

    def log_roc_reading(self, slice: str, threshold: float, tpr: float,
                        fpr: float):
        """Logs a single data point in the ROC Curve of a slice.

        Args:
          slice: String representing slice label.
          threshold: Thresold value for the data point.
          tpr: True positive rate value of the data point.
          fpr: False positive rate value of the data point.
        """

        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].log_roc_reading(threshold, tpr, fpr)
        self._update_metadata(slice)

    def load_roc_readings(self, slice: str, readings: List[List[float]]):
        """Supports bulk loading ROC Curve readings for a slice.

        Args:
          slice: String representing slice label.
          readings: A 2-D list providing ROC Curve data points.
                    The expected order of the data points is: threshold,
                      true_positive_rate, false_positive_rate.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].load_roc_readings(readings)
        self._update_metadata(slice)

    def set_confusion_matrix_categories(self, slice: str,
                                        categories: List[str]):
        """Stores confusion matrix categories for a slice..

        Categories are stored in the internal metrics_utils.ConfusionMatrix
        instance of the slice.

        Args:
          slice: String representing slice label.
          categories: List of strings specifying the categories.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].set_confusion_matrix_categories(categories)
        self._update_metadata(slice)

    def log_confusion_matrix_row(self, slice: str, row_category: str,
                                 row: List[int]):
        """Logs a confusion matrix row for a slice.

        Row is updated on the internal metrics_utils.ConfusionMatrix
        instance of the slice.

        Args:
          slice: String representing slice label.
          row_category: Category to which the row belongs.
          row: List of integers specifying the values for the row.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].log_confusion_matrix_row(row_category, row)
        self._update_metadata(slice)

    def log_confusion_matrix_cell(self, slice: str, row_category: str,
                                  col_category: str, value: int):
        """Logs a confusion matrix cell for a slice..

        Cell is updated on the internal metrics_utils.ConfusionMatrix
        instance of the slice.

        Args:
          slice: String representing slice label.
          row_category: String representing the name of the row category.
          col_category: String representing the name of the column category.
          value: Int value of the cell.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].log_confusion_matrix_cell(
            row_category, col_category, value)
        self._update_metadata(slice)

    def load_confusion_matrix(self, slice: str, categories: List[str],
                              matrix: List[List[int]]):
        """Supports bulk loading the whole confusion matrix for a slice.

        Args:
          slice: String representing slice label.
          categories: List of the category names.
          matrix: Complete confusion matrix.
        """
        self._upsert_classification_metrics_for_slice(slice)
        self._sliced_metrics[slice].log_confusion_matrix_cell(
            categories, matrix)
        self._update_metadata(slice)


class HTML(Artifact):
    """An artifact representing an HTML file."""
    TYPE_NAME = 'system.HTML'

    def __init__(self,
                 name: Optional[str] = None,
                 uri: Optional[str] = None,
                 metadata: Optional[Dict] = None):
        super().__init__(uri=uri, name=name, metadata=metadata)


class Markdown(Artifact):
    """An artifact representing an Markdown file."""
    TYPE_NAME = 'system.Markdown'

    def __init__(self,
                 name: Optional[str] = None,
                 uri: Optional[str] = None,
                 metadata: Optional[Dict] = None):
        super().__init__(uri=uri, name=name, metadata=metadata)


_SCHEMA_TITLE_TO_TYPE: Dict[str, Artifact] = {
    x.TYPE_NAME: x
    for x in [Artifact, Model, Dataset, Metrics, ClassificationMetrics]
}


def create_runtime_artifact(runtime_artifact: Dict) -> Artifact:
    """Creates an Artifact instance from the specified RuntimeArtifact.

    Args:
      runtime_artifact: Dictionary representing JSON-encoded RuntimeArtifact.
    """
    schema_title = runtime_artifact.get('type', {}).get('schemaTitle', '')

    artifact_type = _SCHEMA_TITLE_TO_TYPE.get(schema_title)
    if not artifact_type:
        artifact_type = Artifact
    return artifact_type(
        uri=runtime_artifact.get('uri', ''),
        name=runtime_artifact.get('name', ''),
        metadata=runtime_artifact.get('metadata', {}),
    )



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions for component spec."""

import dataclasses
import enum
import json
from typing import Any, Dict, Mapping, Optional, Sequence, Union

from kfp.components import _components
from kfp.components import structures
import pydantic
import yaml


class InputSpec(pydantic.BaseModel):
    """Component input definitions.

    Attributes:
      type: The type of the input.
      default: Optional; the default value for the input.
    """
    # TODO(ji-yaqi): Add logic to cast default value into the specified type.
    type: str
    default: Optional[Union[str, int, float, bool, dict, list]] = None


class OutputSpec(pydantic.BaseModel):
    """Component output definitions.

    Attributes:
      type: The type of the output.
    """
    type: Union[str, int, float, bool, dict, list]


class BasePlaceholder(pydantic.BaseModel):
    """Base class for placeholders that could appear in container cmd and args.

    Attributes:
      name: Referencing an input or an output from the component.
    """
    name: str


class InputValuePlaceholder(BasePlaceholder):
    pass


class InputPathPlaceholder(BasePlaceholder):
    pass


class InputUriPlaceholder(BasePlaceholder):
    pass


class OutputPathPlaceholder(BasePlaceholder):
    pass


class OutputUriPlaceholder(BasePlaceholder):
    pass


@dataclasses.dataclass
class ResourceSpec:
    """The resource requirements of a container execution.

    Attributes:
      cpu_limit: Optional; the limit of the number of vCPU cores.
      memory_limit: Optional; the memory limit in GB.
      accelerator_type: Optional; the type of accelerators attached to the
        container.
      accelerator_count: Optional; the number of accelerators attached.
    """
    cpu_limit: Optional[float] = None
    memory_limit: Optional[float] = None
    accelerator_type: Optional[str] = None
    accelerator_count: Optional[int] = None


class ContainerSpec(pydantic.BaseModel):
    """Container implementation definition.

    Attributes:
      image: The container image.
      commands: Optional; the container entrypoint.
      arguments: Optional; the arguments to the container entrypoint.
      env: Optional; the environment variables to be passed to the container.
      resources: Optional; the specification on the resource requirements.
    """
    image: str
    commands: Optional[Sequence[Union[str, BasePlaceholder]]] = None
    arguments: Optional[Sequence[Union[str, BasePlaceholder]]] = None
    env: Optional[Mapping[str, Union[str, BasePlaceholder]]] = None
    resources: Optional[ResourceSpec] = None


@dataclasses.dataclass
class ImporterSpec:
    """ImporterSpec definition.

    Attributes:
      artifact_uri: The URI of the artifact.
      type_schema: The type of the artifact.
      reimport: Whether or not import an artifact regardless it has been imported
        before.
      metadata: Optional; the properties of the artifact.
    """
    artifact_uri: str
    type_schema: str
    reimport: bool
    metadata: Optional[Mapping[str, Any]] = None


@dataclasses.dataclass
class TaskSpec:
    """The spec of a pipeline task.

    Attributes:
      name: The name of the task.
      inputs: The sources of task inputs. Constant values or PipelineParams.
      dependent_tasks: The list of upstream tasks.
      enable_caching: Whether or not to enable caching for the task.
      component_ref: The name of a component spec this task is based on.
      trigger_condition: Optional; an expression which will be evaluated into a
        boolean value. True to trigger the task to run.
      trigger_strategy: Optional; when the task will be ready to be triggered.
        Valid values include: "TRIGGER_STRATEGY_UNSPECIFIED",
          "ALL_UPSTREAM_TASKS_SUCCEEDED", and "ALL_UPSTREAM_TASKS_COMPLETED".
      iterator_items: Optional; the items to iterate on. A constant value or a
        PipelineParam.
      iterator_item_input: Optional; the name of the input which has the item from
        the [items][] collection.
    """
    name: str
    inputs: Mapping[str, Any]
    dependent_tasks: Sequence[str]
    enable_caching: bool
    component_ref: str
    trigger_condition: Optional[str] = None
    trigger_strategy: Optional[str] = None
    iterator_items: Optional[Any] = None
    iterator_item_input: Optional[str] = None


@dataclasses.dataclass
class DagSpec:
    """DAG(graph) implementation definition.

    Attributes:
      tasks: The tasks inside the DAG.
      outputs: Defines how the outputs of the dag are linked to the sub tasks.
    """
    tasks: Mapping[str, TaskSpec]
    # TODO(chensun): revisit if we need a DagOutputsSpec class.
    outputs: Mapping[str, Any]


class SchemaVersion(str, enum.Enum):
    V1 = 'v1'
    V2 = 'v2'


class ComponentSpec(pydantic.BaseModel):
    """The definition of a component.

    Attributes:
      name: The name of the component.
      implementation: The implementation of the component. Either an executor
        (container, importer) or a DAG consists of other components.
      inputs: Optional; the input definitions of the component.
      outputs: Optional; the output definitions of the component.
      description: Optional; the description of the component.
      annotations: Optional; the annotations of the component as key-value pairs.
      labels: Optional; the labels of the component as key-value pairs.
      schema_version: Internal field for tracking component version.
    """

    name: str
    description: Optional[str] = None
    implementation: Union[ContainerSpec, ImporterSpec, DagSpec]
    inputs: Optional[Dict[str, InputSpec]] = None
    outputs: Optional[Dict[str, OutputSpec]] = None
    description: Optional[str] = None
    annotations: Optional[Mapping[str, str]] = None
    labels: Optional[Mapping[str, str]] = None
    schema_version: SchemaVersion = SchemaVersion.V2

    def _validate_placeholders(
        self,
        implementation: Union[ContainerSpec, ImporterSpec, DagSpec],
    ) -> Union[ContainerSpec, ImporterSpec, DagSpec]:
        """Validates placeholders reference existing input/output names.

        Args:
          implementation: The component implementation spec.

        Returns:
          The original component implementation spec if no validation error.

        Raises:
          ValueError: if any placeholder references a non-existing input or output.
          TypeError: if any argument is neither a str nor a placeholder instance.
        """
        if not isinstance(implementation, ContainerSpec):
            return implementation

        input_names = [input_spec.name for input_spec in self.input_specs or []]
        output_names = [
            output_spec.name for output_spec in self.output_specs or []
        ]

        for arg in [
                *(implementation.commands or []),
                *(implementation.arguments or [])
        ]:
            if isinstance(arg, (InputValuePlaceholder, InputPathPlaceholder,
                                InputUriPlaceholder)):
                if arg.name not in input_names:
                    raise ValueError(
                        f'Argument "{arg}" references non-existing input.')
            elif isinstance(arg, (OutputPathPlaceholder, OutputUriPlaceholder)):
                if arg.name not in output_names:
                    raise ValueError(
                        f'Argument "{arg}" references non-existing output.')
            # TODO(chensun): revisit if we need to support IfPlaceholder,
            # ConcatPlaceholder, etc. in the new format.
            elif not isinstance(arg, str):
                raise TypeError(f'Unexpected argument "{arg}".')
        return implementation

    @classmethod
    def from_v1_component_spec(
            cls,
            v1_component_spec: structures.ComponentSpec) -> 'ComponentSpec':
        """Converts V1 ComponentSpec to V2 ComponentSpec.

        Args:
          v1_component_spec: The V1 ComponentSpec.

        Returns:
          Component spec in the form of V2 ComponentSpec.

        Raises:
          ValueError: If implementation is not found.
          TypeError: if any argument is neither a str nor Dict.
        """
        component_dict = v1_component_spec.to_dict()
        if component_dict.get('implementation') is None:
            raise ValueError('Implementation field not found')
        if 'container' not in component_dict.get('implementation'):
            raise NotImplementedError

        def _transform_arg(
                arg: Union[str, Dict[str, str]]) -> Union[str, BasePlaceholder]:
            if isinstance(arg, str):
                return arg
            elif 'inputValue' in arg:
                return InputValuePlaceholder(name=arg['inputValue'])
            elif 'inputPath' in arg:
                return InputPathPlaceholder(name=arg['inputPath'])
            elif 'inputUri' in arg:
                return InputUriPlaceholder(name=arg['inputUri'])
            elif 'outputPath' in arg:
                return OutputPathPlaceholder(name=arg['outputPath'])
            elif 'outputUri' in arg:
                return OutputUriPlaceholder(name=arg['outputUri'])
            else:
                raise ValueError(
                    f'Unexpected command/argument type: "{arg}" of type "{type(arg)}".'
                )

        implementation = component_dict['implementation']['container']
        implementation['commands'] = [
            _transform_arg(command)
            for command in implementation.pop('command', [])
        ]
        implementation['arguments'] = [
            _transform_arg(command)
            for command in implementation.pop('args', [])
        ]
        implementation['env'] = {
            key: _transform_arg(command)
            for key, command in implementation.pop('env', {}).items()
        }
        container_spec = ContainerSpec.parse_obj(implementation)

        return ComponentSpec(
            name=component_dict.get('name', 'name'),
            description=component_dict.get('description'),
            implementation=container_spec,
            inputs={
                spec['name']: InputSpec(
                    type=spec.get('type', 'Artifact'),
                    default=spec.get('default', None))
                for spec in component_dict.get('inputs', [])
            },
            outputs={
                spec['name']: OutputSpec(type=spec.get('type', 'String'))
                for spec in component_dict.get('outputs', [])
            },
            schema_version=SchemaVersion.V1)

    def to_v1_component_spec(self) -> structures.ComponentSpec:
        """Converts to v1 ComponentSpec.

        Returns:
          Component spec in the form of V1 ComponentSpec.

        Needed until downstream accept new ComponentSpec.
        """
        if isinstance(self.implementation, DagSpec):
            raise NotImplementedError

        def _transform_arg(arg: Union[str, BasePlaceholder]) -> Any:
            if isinstance(arg, str):
                return arg
            elif isinstance(arg, InputValuePlaceholder):
                return structures.InputValuePlaceholder(arg.name)
            elif isinstance(arg, InputPathPlaceholder):
                return structures.InputPathPlaceholder(arg.name)
            elif isinstance(arg, InputUriPlaceholder):
                return structures.InputUriPlaceholder(arg.name)
            elif isinstance(arg, OutputPathPlaceholder):
                return structures.OutputPathPlaceholder(arg.name)
            elif isinstance(arg, OutputUriPlaceholder):
                return structures.OutputUriPlaceholder(arg.name)
            else:
                # TODO(chensun): transform additional placeholders: if, concat, etc.?
                raise ValueError(
                    f'Unexpected command/argument type: "{arg}" of type "{type(arg)}".'
                )

        return structures.ComponentSpec(
            name=self.name,
            inputs=[
                structures.InputSpec(
                    name=name,
                    type=input_spec.type,
                    default=input_spec.default,
                ) for name, input_spec in self.inputs.items()
            ],
            outputs=[
                structures.OutputSpec(
                    name=name,
                    type=output_spec.type,
                ) for name, output_spec in self.outputs.items()
            ],
            implementation=structures.ContainerImplementation(
                container=structures.ContainerSpec(
                    image=self.implementation.image,
                    command=[
                        _transform_arg(cmd)
                        for cmd in self.implementation.commands or []
                    ],
                    args=[
                        _transform_arg(arg)
                        for arg in self.implementation.arguments or []
                    ],
                    env={
                        name: _transform_arg(value)
                        for name, value in self.implementation.env or {}
                    },
                )),
        )

    @classmethod
    def load_from_component_yaml(cls, component_yaml: str) -> 'ComponentSpec':
        """Loads V1 or V2 component yaml into ComponentSpec.

        Args:
          component_yaml: the component yaml in string format.

        Returns:
          Component spec in the form of V2 ComponentSpec.
        """

        json_component = yaml.safe_load(component_yaml)

        if 'schema_version' in json_component and json_component[
                'schema_version'] == SchemaVersion.V2:
            return ComponentSpec.parse_obj(json_component)

        v1_component = _components._load_component_spec_from_component_text(
            component_yaml)
        return cls.from_v1_component_spec(v1_component)

    def save_to_component_yaml(self, output_file: str) -> None:
        """Saves ComponentSpec into yaml file.

        Args:
          output_file: File path to store the component yaml.
        """
        with open(output_file, 'a') as output_file:
            json_component = self.json(exclude_none=True)
            yaml_file = yaml.safe_dump(json.loads(json_component))
            output_file.write(yaml_file)



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.compiler.compiler_utils."""

import unittest

from kfp.dsl import _pipeline_param
from kfp.v2.compiler import compiler_utils
from kfp.pipeline_spec import pipeline_spec_pb2
from google.protobuf import json_format
from google.protobuf import message


class CompilerUtilsTest(unittest.TestCase):

    def assertProtoEquals(self, proto1: message.Message,
                          proto2: message.Message):
        """Asserts the equality between two messages."""
        self.assertDictEqual(
            json_format.MessageToDict(proto1),
            json_format.MessageToDict(proto2))

    def test_build_runtime_config_spec(self):
        expected_dict = {
            'gcsOutputDirectory': 'gs://path',
            'parameters': {
                'input1': {
                    'stringValue': 'test'
                },
                'input2': {
                    'intValue': 2
                },
                'input3': {
                    'stringValue': '[1, 2, 3]'
                }
            }
        }
        expected_spec = pipeline_spec_pb2.PipelineJob.RuntimeConfig()
        json_format.ParseDict(expected_dict, expected_spec)

        runtime_config = compiler_utils.build_runtime_config_spec(
            'gs://path', {
                'input1':
                    _pipeline_param.PipelineParam(
                        name='input1', param_type='String', value='test'),
                'input2':
                    _pipeline_param.PipelineParam(
                        name='input2', param_type='Integer', value=2),
                'input3':
                    _pipeline_param.PipelineParam(
                        name='input3', param_type='List', value=[1, 2, 3]),
                'input4':
                    _pipeline_param.PipelineParam(
                        name='input4', param_type='Double', value=None)
            })
        self.assertEqual(expected_spec, runtime_config)

    def test_validate_pipeline_name(self):
        compiler_utils.validate_pipeline_name('my-pipeline')

        compiler_utils.validate_pipeline_name('p' * 128)

        with self.assertRaisesRegex(ValueError, 'Invalid pipeline name: '):
            compiler_utils.validate_pipeline_name('my_pipeline')

        with self.assertRaisesRegex(ValueError, 'Invalid pipeline name: '):
            compiler_utils.validate_pipeline_name('My pipeline')

        with self.assertRaisesRegex(ValueError, 'Invalid pipeline name: '):
            compiler_utils.validate_pipeline_name('-my-pipeline')

        with self.assertRaisesRegex(ValueError, 'Invalid pipeline name: '):
            compiler_utils.validate_pipeline_name('p' * 129)

    def test_refactor_v2_component_success(self):
        test_v2_container_spec = compiler_utils.PipelineContainerSpec(
            image='my/dummy-image',
            command=['python', '-m', 'my_package.my_entrypoint'],
            args=['arg1', 'arg2', '--function_name', 'test_func'])
        expected_container_spec = compiler_utils.PipelineContainerSpec(
            image='my/dummy-image',
            command=['python', '-m', 'kfp.container.entrypoint'],
            args=[
                '--executor_input_str', '{{$}}', '--function_name', 'test_func'
            ])
        compiler_utils.refactor_v2_container_spec(test_v2_container_spec)
        self.assertProtoEquals(expected_container_spec, test_v2_container_spec)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.google.client.client."""

import datetime
import json
import os
import unittest
from typing import Any, Dict
from unittest import mock

from googleapiclient import discovery, http
from kfp.v2.google.client import client, client_utils

# Mock response for get job request.
_EXPECTED_GET_RESPONSE = 'good job spec'
_GET_RESPONSES = {
    'projects/test-project/locations/us-central1/'
    'pipelineJobs/test-job':
        _EXPECTED_GET_RESPONSE
}


def _load_test_data(filename: str) -> Dict[Any, Any]:
    """Test helper function that loads json from testdata."""
    file_path = os.path.join(os.path.dirname(__file__), 'testdata', filename)
    with open(file_path) as f:
        return json.load(f)


# Mock the Python API client classes for testing purpose.
class _MockClient(object):
    """Mocks Python Google API client."""

    def projects(self):  # pylint: disable=invalid-name
        return _MockProjectsResource()


class _MockProjectsResource(object):
    """Mocks API Resource returned by projects()."""

    def locations(self):  # pylint: disable=invalid-name
        return _MockLocationResource()


class _MockLocationResource(object):
    """Mocks API Resource returned by locations()."""

    def pipelineJobs(self):  # pylint: disable=invalid-name
        return _MockPipelineJobsResource()


class _MockPipelineJobsResource(object):
    """Mocks API Resource returned by pipelineJobs()."""

    class _MockGetRequest(http.HttpRequest):

        def __init__(self, name: str):
            self._name = name

        def execute(self):
            return _GET_RESPONSES.get(self._name)

    def get(self, name: str):  # pylint: disable=invalid-name
        """Mocks get job request."""
        return self._MockGetRequest(name=name)


class ClientTest(unittest.TestCase):

    def test_client_init_with_defaults(self):
        api_client = client.AIPlatformClient(
            project_id='test-project', region='us-central1')

        self.assertEqual(api_client._project_id, 'test-project')
        self.assertEqual(api_client._region, 'us-central1')
        self.assertEqual(api_client._endpoint,
                         'us-central1-aiplatform.googleapis.com')

    @mock.patch.object(client, '_get_current_time', autospec=True)
    @mock.patch.object(client_utils, 'load_json', autospec=True)
    def test_create_run_from_pipeline_job(self, mock_load_json,
                                          mock_get_current_time):
        mock_load_json.return_value = _load_test_data('pipeline_job.json')
        mock_get_current_time.return_value = datetime.date(2020, 10, 28)

        api_client = client.AIPlatformClient(
            project_id='test-project', region='us-central1')
        with mock.patch.object(
                api_client, '_submit_job', autospec=True) as mock_submit:
            api_client.create_run_from_job_spec(
                job_spec_path='path/to/pipeline_job.json')

            golden = _load_test_data('pipeline_job.json')
            mock_submit.assert_called_once_with(
                job_spec=golden, job_id='sample-test-pipeline-20201028000000')

    @mock.patch.object(client, '_get_current_time', autospec=True)
    @mock.patch.object(client_utils, 'load_json', autospec=True)
    def test_job_id_parameters_override(self, mock_load_json,
                                        mock_get_current_time):
        mock_load_json.return_value = _load_test_data('pipeline_job.json')
        mock_get_current_time.return_value = datetime.date(2020, 10, 28)

        api_client = client.AIPlatformClient(
            project_id='test-project', region='us-central1')
        with mock.patch.object(
                api_client, '_submit_job', autospec=True) as mock_submit:
            api_client.create_run_from_job_spec(
                job_spec_path='path/to/pipeline_job.json',
                job_id='my-new-id',
                pipeline_root='gs://bucket/new-blob',
                parameter_values={
                    'text': 'Hello test!',
                    'list': [1, 2, 3],
                })

            golden = _load_test_data('pipeline_job.json')
            golden['name'] = ('projects/test-project/locations/us-central1/'
                              'pipelineJobs/my-new-id')
            golden['displayName'] = 'my-new-id'
            golden['runtimeConfig'][
                'gcsOutputDirectory'] = 'gs://bucket/new-blob'
            golden['runtimeConfig']['parameters']['text'] = {
                'stringValue': 'Hello test!'
            }
            golden['runtimeConfig']['parameters']['list'] = {
                'stringValue': '[1, 2, 3]'
            }
            mock_submit.assert_called_once_with(
                job_spec=golden, job_id='my-new-id')

    @mock.patch.object(client, '_get_current_time', autospec=True)
    @mock.patch.object(client_utils, 'load_json', autospec=True)
    def test_advanced_settings(self, mock_load_json, mock_get_current_time):
        mock_load_json.return_value = _load_test_data('pipeline_job.json')
        mock_get_current_time.return_value = datetime.date(2020, 10, 28)

        api_client = client.AIPlatformClient(
            project_id='test-project', region='us-central1')
        with mock.patch.object(
                api_client, '_submit_job', autospec=True) as mock_submit:
            api_client.create_run_from_job_spec(
                job_spec_path='path/to/pipeline_job.json',
                cmek='custom-key',
                service_account='custom-sa',
                network='custom-network')

            golden = _load_test_data('pipeline_job.json')
            golden['encryptionSpec'] = {'kmsKeyName': 'custom-key'}
            golden['serviceAccount'] = 'custom-sa'
            golden['network'] = 'custom-network'
            mock_submit.assert_called_once_with(
                job_spec=golden, job_id='sample-test-pipeline-20201028000000')

    def test_extract_job_id_success(self):

        job_name = 'projects/123/locations/us-central1/pipelineJobs/0123456789'
        self.assertEqual('0123456789', client._extract_job_id(job_name))
        self.assertIsNone(client._extract_job_id('invalid name'))

    @mock.patch.object(client, '_get_current_time', autospec=True)
    @mock.patch.object(client, '_get_gcp_credential', autospec=True)
    @mock.patch.object(discovery, 'build_from_document', autospec=True)
    def test_get_job_success(self, mock_build_client, mock_get_gcp_credential,
                             mock_get_current_time):
        mock_get_current_time.return_value = datetime.date(2020, 10, 28)
        mock_build_client.return_value = _MockClient()
        api_client = client.AIPlatformClient(
            project_id='test-project', region='us-central1')

        self.assertEqual(_EXPECTED_GET_RESPONSE,
                         api_client.get_job(job_id='test-job'))
        mock_get_gcp_credential.assert_called_once()

    @mock.patch.object(client, '_get_current_time', autospec=True)
    @mock.patch.object(client_utils, 'load_json', autospec=True)
    def test_disable_caching(self, mock_load_json, mock_get_current_time):
        mock_load_json.return_value = _load_test_data('pipeline_job.json')
        mock_get_current_time.return_value = datetime.date(2020, 10, 28)

        api_client = client.AIPlatformClient(
            project_id='test-project', region='us-central1')
        with mock.patch.object(
                api_client, '_submit_job', autospec=True) as mock_submit:
            api_client.create_run_from_job_spec(
                job_spec_path='path/to/pipeline_job.json', enable_caching=False)

            golden = _load_test_data('pipeline_job.json')
            golden = json.loads(
                json.dumps(golden).replace('"enableCache": true',
                                           '"enableCache": false'))
            mock_submit.assert_called_once_with(
                job_spec=golden, job_id='sample-test-pipeline-20201028000000')

    @mock.patch.object(client, '_get_current_time', autospec=True)
    @mock.patch.object(client_utils, 'load_json', autospec=True)
    def test_setting_labels(self, mock_load_json, mock_get_current_time):
        mock_load_json.return_value = _load_test_data('pipeline_job.json')
        mock_get_current_time.return_value = datetime.date(2020, 10, 28)

        user_labels = {'label1': 'value1', 'label2': 'value2'}
        api_client = client.AIPlatformClient(
            project_id='test-project', region='us-central1')
        with mock.patch.object(
                api_client, '_submit_job', autospec=True) as mock_submit:
            api_client.create_run_from_job_spec(
                job_spec_path='path/to/pipeline_job.json', labels=user_labels)

            golden = _load_test_data('pipeline_job.json')
            golden['labels'] = user_labels
            mock_submit.assert_called_once_with(
                job_spec=golden, job_id='sample-test-pipeline-20201028000000')


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""KFP DSL v2 compiler.

This is an experimental implementation of KFP compiler that compiles KFP
pipeline into Pipeline IR:
https://docs.google.com/document/d/1PUDuSQ8vmeKSBloli53mp7GIvzekaY7sggg6ywy35Dk/
"""

import collections
import inspect
import json
import uuid
import warnings
from typing import (Any, Callable, Dict, List, Mapping, Optional, Set, Tuple,
                    Union)

import kfp
from google.protobuf import json_format
from kfp import dsl
from kfp.compiler._k8s_helper import sanitize_k8s_name
from kfp.dsl import _for_loop
from kfp.dsl import component_spec as dsl_component_spec
from kfp.dsl import dsl_utils
from kfp.pipeline_spec import pipeline_spec_pb2
from kfp.v2.compiler import compiler_utils
from kfp.v2.components import component_factory
from kfp.v2.components.types import artifact_types, type_utils

_GroupOrOp = Union[dsl.OpsGroup, dsl.BaseOp]


class Compiler(object):
    """Experimental DSL compiler that targets the PipelineSpec IR.

    It compiles pipeline function into PipelineSpec json string.
    PipelineSpec is the IR protobuf message that defines a pipeline:
    https://github.com/kubeflow/pipelines/blob/237795539f7b85bac77435e2464367226ee19391/api/v2alpha1/pipeline_spec.proto#L8
    In this initial implementation, we only support components authored through
    Component yaml spec. And we don't support advanced features like conditions,
    static and dynamic loops, etc.

    Example:
      How to use the compiler to construct pipeline_spec json:

        @dsl.pipeline(
          name='name',
          description='description'
        )
        def my_pipeline(a: int = 1, b: str = "default value"):
          ...

        kfp.v2.compiler.Compiler().compile(my_pipeline, 'path/to/pipeline.json')
    """

    def _get_groups_for_ops(self,
                            root_group: dsl.OpsGroup) -> Dict[str, List[str]]:
        """Helper function to get groups that contain the specified ops.

        Each pipeline has a root group. Each group has a list of operators (leaf)
        and groups.
        This function traverse the tree and get all ancestor groups for all
        operators.

        Args:
         root_group: The root node of a ops tree or subtree.

        Returns:
          A dict. Key is the operator's name. Value is a list of ancestor groups
          including the op itself. The list of a given operator is sorted in a way
          that the farthest group is the first and operator itself is the last.
        """

        def _get_op_groups_helper(
                current_groups: List[dsl.OpsGroup],
                ops_to_groups: Dict[str, List[dsl.OpsGroup]]) -> None:
            root_group = current_groups[-1]
            for g in root_group.groups:
                # Add recursive opsgroup in the ops_to_groups
                # such that the i/o dependency can be propagated to the ancester opsgroups
                if g.recursive_ref:
                    ops_to_groups[g.name] = [x.name for x in current_groups
                                            ] + [g.name]
                    continue
                current_groups.append(g)
                _get_op_groups_helper(current_groups, ops_to_groups)
                del current_groups[-1]
            for op in root_group.ops:
                ops_to_groups[op.name] = [x.name for x in current_groups
                                         ] + [op.name]

        ops_to_groups = {}
        current_groups = [root_group]
        _get_op_groups_helper(current_groups, ops_to_groups)
        return ops_to_groups

    #TODO: combine with the _get_groups_for_ops
    def _get_groups_for_opsgroups(
            self, root_group: dsl.OpsGroup) -> Dict[str, List[str]]:
        """Helper function to get groups that contain the specified opsgroup.

        Each pipeline has a root group. Each group has a list of operators (leaf)
        and groups.
        This function traverse the tree and get all ancestor groups for all
        opsgroups.

        Args:
         root_group: The root node of a groups tree or subtree.

        Returns:
          A dict. Key is the opsgroup's name. Value is a list of ancestor groups
          including the opsgroup itself. The list of a given opsgroup is sorted in a
          way that the farthest group is the first and opsgroup itself is the last.
        """

        def _get_opsgroup_groups_helper(
                current_groups: dsl.OpsGroup,
                opsgroups_to_groups: Dict[str, List[dsl.OpsGroup]]) -> None:
            root_group = current_groups[-1]
            for g in root_group.groups:
                # Add recursive opsgroup in the ops_to_groups
                # such that the i/o dependency can be propagated to the ancester opsgroups
                if g.recursive_ref:
                    continue
                opsgroups_to_groups[g.name] = [x.name for x in current_groups
                                              ] + [g.name]
                current_groups.append(g)
                _get_opsgroup_groups_helper(current_groups, opsgroups_to_groups)
                del current_groups[-1]

        opsgroups_to_groups = {}
        current_groups = [root_group]
        _get_opsgroup_groups_helper(current_groups, opsgroups_to_groups)
        return opsgroups_to_groups

    def _get_groups(self, root_group: dsl.OpsGroup) -> Dict[str, dsl.OpsGroup]:
        """Helper function to get all groups (not including ops) in a
        pipeline."""

        def _get_groups_helper(group):
            groups = {group.name: group}
            for g in group.groups:
                # Skip the recursive opsgroup because no templates
                # need to be generated for the recursive opsgroups.
                if not g.recursive_ref:
                    groups.update(_get_groups_helper(g))
            return groups

        return _get_groups_helper(root_group)

    def _get_uncommon_ancestors(
        self,
        op_groups: Dict[str, List[str]],
        opsgroup_groups: Dict[str, List[str]],
        op1: dsl.BaseOp,
        op2: dsl.BaseOp,
    ) -> Tuple[List[_GroupOrOp], List[_GroupOrOp]]:
        """Helper function to get unique ancestors between two ops.

        For example, op1's ancestor groups are [root, G1, G2, G3, op1],
        op2's ancestor groups are [root, G1, G4, op2], then it returns a
        tuple ([G2, G3, op1], [G4, op2]).
        """
        #TODO: extract a function for the following two code module
        if op1.name in op_groups:
            op1_groups = op_groups[op1.name]
        elif op1.name in opsgroup_groups:
            op1_groups = opsgroup_groups[op1.name]
        else:
            raise ValueError(op1.name + ' does not exist.')

        if op2.name in op_groups:
            op2_groups = op_groups[op2.name]
        elif op2.name in opsgroup_groups:
            op2_groups = opsgroup_groups[op2.name]
        else:
            raise ValueError(op2.name + ' does not exist.')

        both_groups = [op1_groups, op2_groups]
        common_groups_len = sum(
            1 for x in zip(*both_groups) if x == (x[0],) * len(x))
        group1 = op1_groups[common_groups_len:]
        group2 = op2_groups[common_groups_len:]
        return (group1, group2)

    def _get_condition_params_for_ops(
            self, root_group: dsl.OpsGroup) -> Dict[str, dsl.PipelineParam]:
        """Get parameters referenced in conditions of ops."""
        conditions = collections.defaultdict(set)

        def _get_condition_params_for_ops_helper(group,
                                                 current_conditions_params):
            new_current_conditions_params = current_conditions_params
            if group.type == 'condition':
                new_current_conditions_params = list(current_conditions_params)
                if isinstance(group.condition.operand1, dsl.PipelineParam):
                    new_current_conditions_params.append(
                        group.condition.operand1)
                if isinstance(group.condition.operand2, dsl.PipelineParam):
                    new_current_conditions_params.append(
                        group.condition.operand2)
            for op in group.ops:
                for param in new_current_conditions_params:
                    conditions[op.name].add(param)
            for g in group.groups:
                # If the subgroup is a recursive opsgroup, propagate the pipelineparams
                # in the condition expression, similar to the ops.
                if g.recursive_ref:
                    for param in new_current_conditions_params:
                        conditions[g.name].add(param)
                else:
                    _get_condition_params_for_ops_helper(
                        g, new_current_conditions_params)

        _get_condition_params_for_ops_helper(root_group, [])
        return conditions

    def _get_next_group_or_op(self, to_visit: List, already_visited: Set):
        """Get next group or op to visit."""
        if len(to_visit) == 0:
            return None
        next = to_visit.pop(0)
        while next in already_visited:
            next = to_visit.pop(0)
        already_visited.add(next)
        return next

    def _get_for_loop_ops(self, new_root) -> Dict[str, dsl.ParallelFor]:
        to_visit = self._get_all_subgroups_and_ops(new_root)
        op_name_to_op = {}
        already_visited = set()

        while len(to_visit):
            next_op = self._get_next_group_or_op(to_visit, already_visited)
            if next_op is None:
                break
            to_visit.extend(self._get_all_subgroups_and_ops(next_op))
            if isinstance(next_op, dsl.ParallelFor):
                op_name_to_op[next_op.name] = next_op

        return op_name_to_op

    def _get_all_subgroups_and_ops(self, group: dsl.OpsGroup):
        """Get all ops and groups contained within this group."""
        subgroups = []
        if hasattr(group, 'ops'):
            subgroups.extend(group.ops)
        if hasattr(group, 'groups'):
            subgroups.extend(group.groups)
        return subgroups

    def _get_inputs_outputs(
        self,
        pipeline: dsl.Pipeline,
        args: List[dsl.PipelineParam],
        root_group: dsl.OpsGroup,
        op_groups: Dict[str, List[str]],
        opsgroup_groups: Dict[str, List[str]],
        condition_params: Dict[str, dsl.PipelineParam],
        op_name_to_for_loop_op: Dict[str, dsl.ParallelFor],
    ) -> Tuple[Dict[str, List[Tuple[dsl.PipelineParam, str]]], Dict[
            str, List[Tuple[dsl.PipelineParam, str]]]]:
        """Get inputs and outputs of each group and op.

        Args:
          pipeline: The instantiated pipeline object.
          args: The list of pipeline function arguments as PipelineParam.
          root_group: The root OpsGroup.
          op_groups: The dict of op name to parent groups.
          opsgroup_groups: The dict of opsgroup name to parent groups.
          condition_params: The dict of group name to pipeline params referenced in
            the conditions in that group.
          op_name_to_for_loop_op: The dict of op name to loop ops.

        Returns:
          A tuple (inputs, outputs).
          inputs and outputs are dicts with key being the group/op names and values
          being list of tuples (param, producing_op_name). producing_op_name is the
          name of the op that produces the param. If the param is a pipeline param
          (no producer op), then producing_op_name is None.
        """
        inputs = collections.defaultdict(set)
        outputs = collections.defaultdict(set)

        # Fix possible missing type -- PipelineParam parsed from command line args
        # doesn't contain the type information, as the `.param_type` is not included
        # during PipelineParam serialization.
        all_params = {param.pattern: param for param in args}
        for op in pipeline.ops.values():
            for param in op.inputs + list(op.outputs.values()) + list(
                    condition_params[op.name]):
                if param.pattern not in all_params:
                    all_params[param.pattern] = param
                else:
                    param.param_type = param.param_type or all_params[
                        param.pattern].param_type
                    all_params[param.pattern].param_type = param.param_type

        for op in pipeline.ops.values():
            # op's inputs and all params used in conditions for that op are both
            # considered.
            for param in op.inputs + list(condition_params[op.name]):

                # if the value is already provided (immediate value), then no need to
                # expose it as input for its parent groups.
                if param.value:
                    continue
                if param.op_name:
                    upstream_op = pipeline.ops[param.op_name]
                    upstream_groups, downstream_groups = (
                        self._get_uncommon_ancestors(op_groups, opsgroup_groups,
                                                     upstream_op, op))
                    for i, group_name in enumerate(downstream_groups):
                        if i == 0:
                            # If it is the first uncommon downstream group, then the input
                            # comes from the first uncommon upstream group.
                            inputs[group_name].add((param, upstream_groups[0]))
                        else:
                            # If not the first downstream group, then the input is passed down
                            # from its ancestor groups so the upstream group is None.
                            inputs[group_name].add((param, None))
                    for i, group_name in enumerate(upstream_groups):
                        if i == len(upstream_groups) - 1:
                            # If last upstream group, it is an operator and output comes from container.
                            outputs[group_name].add((param, None))
                        else:
                            # If not last upstream group, output value comes from one of its child.
                            outputs[group_name].add(
                                (param, upstream_groups[i + 1]))
                else:
                    if not op.is_exit_handler:
                        for group_name in op_groups[op.name][::-1]:
                            # if group is for loop group and param is that loop's param, then the param
                            # is created by that for loop ops_group and it shouldn't be an input to
                            # any of its parent groups.
                            inputs[group_name].add((param, None))
                            if group_name in op_name_to_for_loop_op:
                                # for example:
                                #   loop_group.loop_args.name = 'loop-item-param-99ca152e'
                                #   param.name =                'loop-item-param-99ca152e--a'
                                loop_group = op_name_to_for_loop_op[group_name]
                                if loop_group.loop_args.name in param.name:
                                    break

        # Generate the input/output for recursive opsgroups
        # It propagates the recursive opsgroups IO to their ancester opsgroups
        def _get_inputs_outputs_recursive_opsgroup(group: dsl.OpsGroup):
            #TODO: refactor the following codes with the above
            if group.recursive_ref:
                params = [(param, False) for param in group.inputs]
                params.extend([(param, True)
                               for param in list(condition_params[group.name])])
                for param, is_condition_param in params:
                    if param.value:
                        continue

                    if param.op_name:
                        upstream_op = pipeline.ops[param.op_name]
                        upstream_groups, downstream_groups = \
                          self._get_uncommon_ancestors(op_groups, opsgroup_groups, upstream_op, group)
                        for i, g in enumerate(downstream_groups):
                            if i == 0:
                                inputs[g].add((param, upstream_groups[0]))
                            # There is no need to pass the condition param as argument to the downstream ops.
                            #TODO: this might also apply to ops. add a TODO here and think about it.
                            elif i == len(downstream_groups
                                         ) - 1 and is_condition_param:
                                continue
                            else:
                                inputs[g].add((param, None))
                        for i, g in enumerate(upstream_groups):
                            if i == len(upstream_groups) - 1:
                                outputs[g].add((param, None))
                            else:
                                outputs[g].add((param, upstream_groups[i + 1]))
                    elif not is_condition_param:
                        for g in op_groups[group.name]:
                            inputs[g].add((param, None))
            for subgroup in group.groups:
                _get_inputs_outputs_recursive_opsgroup(subgroup)

        _get_inputs_outputs_recursive_opsgroup(root_group)

        # Generate the input for SubGraph along with parallelfor
        for subgraph in opsgroup_groups:
            if subgraph in op_name_to_for_loop_op:
                # The opsgroup list is sorted with the farthest group as the first and the opsgroup
                # itself as the last. To get the latest opsgroup which is not the opsgroup itself -2 is used.
                parent = opsgroup_groups[subgraph][-2]
                if parent and parent.startswith('subgraph'):
                    # propagate only op's pipeline param from subgraph to parallelfor
                    loop_op = op_name_to_for_loop_op[subgraph]
                    pipeline_param = loop_op.loop_args.items_or_pipeline_param
                    if loop_op.items_is_pipeline_param and pipeline_param.op_name:
                        inputs[parent].add(
                            (pipeline_param, pipeline_param.op_name))

        return inputs, outputs

    def _get_dependencies(
        self,
        pipeline: dsl.Pipeline,
        root_group: dsl.OpsGroup,
        op_groups: Dict[str, List[str]],
        opsgroups_groups: Dict[str, List[str]],
        opsgroups: Dict[str, dsl.OpsGroup],
        condition_params: Dict[str, dsl.PipelineParam],
    ) -> Dict[str, List[_GroupOrOp]]:
        """Get dependent groups and ops for all ops and groups.

        Args:
          pipeline: The instantiated pipeline object.
          root_group: The root OpsGroup.
          op_groups: The dict of op name to parent groups.
          opsgroup_groups: The dict of opsgroup name to parent groups.
          opsgroups: The dict of opsgroup name to opsgroup.
          condition_params: The dict of group name to pipeline params referenced in
            the conditions in that group.

        Returns:
          A dict. Key is group/op name, value is a list of dependent groups/ops.
          The dependencies are calculated in the following way: if op2 depends on
          op1, and their ancestors are [root, G1, G2, op1] and
          [root, G1, G3, G4, op2], then G3 is dependent on G2. Basically dependency
          only exists in the first uncommon ancesters in their ancesters chain. Only
          sibling groups/ops can have dependencies.
        """
        dependencies = collections.defaultdict(set)
        for op in pipeline.ops.values():
            upstream_op_names = set()
            for param in op.inputs + list(condition_params[op.name]):
                if param.op_name:
                    upstream_op_names.add(param.op_name)
            upstream_op_names |= set(op.dependent_names)

            for upstream_op_name in upstream_op_names:
                # the dependent op could be either a BaseOp or an opsgroup
                if upstream_op_name in pipeline.ops:
                    upstream_op = pipeline.ops[upstream_op_name]
                elif upstream_op_name in opsgroups:
                    upstream_op = opsgroups[upstream_op_name]
                else:
                    raise ValueError('compiler cannot find the ' +
                                     upstream_op_name)

                upstream_groups, downstream_groups = self._get_uncommon_ancestors(
                    op_groups, opsgroups_groups, upstream_op, op)
                dependencies[downstream_groups[0]].add(upstream_groups[0])

        # Generate dependencies based on the recursive opsgroups
        #TODO: refactor the following codes with the above
        def _get_dependency_opsgroup(
                group: dsl.OpsGroup,
                dependencies: Dict[str, List[_GroupOrOp]]) -> None:
            upstream_op_names = set(
                [dependency.name for dependency in group.dependencies])
            if group.recursive_ref:
                for param in group.inputs + list(condition_params[group.name]):
                    if param.op_name:
                        upstream_op_names.add(param.op_name)

            for op_name in upstream_op_names:
                if op_name in pipeline.ops:
                    upstream_op = pipeline.ops[op_name]
                elif op_name in opsgroups:
                    upstream_op = opsgroups[op_name]
                else:
                    raise ValueError('compiler cannot find the ' + op_name)
                upstream_groups, downstream_groups = (
                    self._get_uncommon_ancestors(op_groups, opsgroups_groups,
                                                 upstream_op, group))
                dependencies[downstream_groups[0]].add(upstream_groups[0])

            for subgroup in group.groups:
                _get_dependency_opsgroup(subgroup, dependencies)

        _get_dependency_opsgroup(root_group, dependencies)

        return dependencies

    def _resolve_condition_operands(
            self, operand1: Union[str, dsl.PipelineParam],
            operand2: Union[str, dsl.PipelineParam]) -> Tuple[str, str]:
        """Resolves values and PipelineParams for condition operands."""

        # Pre-scan the operand to get the type of constant value if there's any.
        # The value_type can be used to backfill missing PipelineParam.param_type.
        value_type = None
        for value_or_reference in [operand1, operand2]:
            if isinstance(value_or_reference, dsl.PipelineParam):
                continue
            if isinstance(value_or_reference, float):
                value_type = 'Float'
            elif isinstance(value_or_reference, int):
                value_type = 'Integer'
            else:
                value_type = 'String'

        operand_values = []
        for value_or_reference in [operand1, operand2]:
            if isinstance(value_or_reference, dsl.PipelineParam):
                input_name = dsl_component_spec.additional_input_name_for_pipelineparam(
                    value_or_reference)
                # Condition operand is always parameters for now.
                value_or_reference.param_type = (
                    value_or_reference.param_type or value_type)
                operand_values.append(
                    "inputs.parameters['{input_name}'].{value_field}".format(
                        input_name=input_name,
                        value_field=type_utils.get_parameter_type_field_name(
                            value_or_reference.param_type)))
            else:
                if isinstance(value_or_reference, str):
                    operand_values.append("'{}'".format(value_or_reference))
                else:
                    operand_values.append(str(value_or_reference))

        return tuple(operand_values)

    def _populate_metrics_in_dag_outputs(
        self,
        ops: List[dsl.ContainerOp],
        op_to_parent_groups: Dict[str, List[str]],
        pipeline_spec: pipeline_spec_pb2.PipelineSpec,
    ) -> None:
        """Populates metrics artifacts in dag outputs.

        Args:
          ops: The list of ops that may produce metrics outputs.
          op_to_parent_groups: The dict of op name to parent groups. Key is the op's
            name. Value is a list of ancestor groups including the op itself. The
            list of a given op is sorted in a way that the farthest group is the
            first and the op itself is the last.
          pipeline_spec: The pipeline_spec to update in-place.
        """
        for op in ops:
            op_task_spec = getattr(op, 'task_spec',
                                   pipeline_spec_pb2.PipelineTaskSpec())
            op_component_spec = getattr(op, 'component_spec',
                                        pipeline_spec_pb2.ComponentSpec())

            # Get the tuple of (component_name, task_name) of all its parent groups.
            parent_components_and_tasks = [('_root', '')]
            # skip the op itself and the root group which cannot be retrived via name.
            for group_name in op_to_parent_groups[op.name][1:-1]:
                parent_components_and_tasks.append(
                    (dsl_utils.sanitize_component_name(group_name),
                     dsl_utils.sanitize_task_name(group_name)))
            # Reverse the order to make the farthest group in the end.
            parent_components_and_tasks.reverse()

            for output_name, artifact_spec in \
                op_component_spec.output_definitions.artifacts.items():

                if artifact_spec.artifact_type.WhichOneof(
                        'kind'
                ) == 'schema_title' and artifact_spec.artifact_type.schema_title in [
                        artifact_types.Metrics.TYPE_NAME,
                        artifact_types.ClassificationMetrics.TYPE_NAME,
                ]:
                    unique_output_name = '{}-{}'.format(op.name, output_name)

                    sub_task_name = op.name
                    sub_task_output = output_name
                    for component_name, task_name in parent_components_and_tasks:
                        group_component_spec = (
                            pipeline_spec.root if component_name == '_root' else
                            pipeline_spec.components[component_name])
                        group_component_spec.output_definitions.artifacts[
                            unique_output_name].CopyFrom(artifact_spec)
                        group_component_spec.dag.outputs.artifacts[
                            unique_output_name].artifact_selectors.append(
                                pipeline_spec_pb2.DagOutputsSpec
                                .ArtifactSelectorSpec(
                                    producer_subtask=sub_task_name,
                                    output_artifact_key=sub_task_output,
                                ))
                        sub_task_name = task_name
                        sub_task_output = unique_output_name

    def _group_to_dag_spec(
        self,
        group: dsl.OpsGroup,
        inputs: Dict[str, List[Tuple[dsl.PipelineParam, str]]],
        outputs: Dict[str, List[Tuple[dsl.PipelineParam, str]]],
        dependencies: Dict[str, List[_GroupOrOp]],
        pipeline_spec: pipeline_spec_pb2.PipelineSpec,
        deployment_config: pipeline_spec_pb2.PipelineDeploymentConfig,
        rootgroup_name: str,
        op_to_parent_groups: Dict[str, List[str]],
    ) -> None:
        """Generate IR spec given an OpsGroup.

        Args:
          group: The OpsGroup to generate spec for.
          inputs: The inputs dictionary. The keys are group/op names and values are
            lists of tuples (param, producing_op_name).
          outputs: The outputs dictionary. The keys are group/op names and values
            are lists of tuples (param, producing_op_name).
          dependencies: The group dependencies dictionary. The keys are group/op
            names, and the values are lists of dependent groups/ops.
          pipeline_spec: The pipeline_spec to update in-place.
          deployment_config: The deployment_config to hold all executors.
          rootgroup_name: The name of the group root. Used to determine whether the
            component spec for the current group should be the root dag.
          op_to_parent_groups: The dict of op name to parent groups. Key is the op's
            name. Value is a list of ancestor groups including the op itself. The
            list of a given op is sorted in a way that the farthest group is the
            first and the op itself is the last.
        """
        group_component_name = dsl_utils.sanitize_component_name(group.name)

        if group.name == rootgroup_name:
            group_component_spec = pipeline_spec.root
        else:
            group_component_spec = pipeline_spec.components[
                group_component_name]

        # Generate task specs and component specs for the dag.
        subgroups = group.groups + group.ops
        for subgroup in subgroups:
            subgroup_task_spec = getattr(subgroup, 'task_spec',
                                         pipeline_spec_pb2.PipelineTaskSpec())
            subgroup_component_spec = getattr(subgroup, 'component_spec',
                                              pipeline_spec_pb2.ComponentSpec())

            display_name = getattr(subgroup, 'display_name', None)
            subgroup_task_spec.task_info.name = display_name or subgroup.name

            is_recursive_subgroup = (
                isinstance(subgroup, dsl.OpsGroup) and subgroup.recursive_ref)
            if is_recursive_subgroup:
                raise NotImplementedError(
                    'Recursive subgroup is not supported in v2 yet.')
            else:
                subgroup_key = subgroup.name

            # human_name exists for ops only, and is used to de-dupe component spec.
            subgroup_component_name = (
                subgroup_task_spec.component_ref.name or
                dsl_utils.sanitize_component_name(
                    getattr(subgroup, 'human_name', subgroup_key)))
            subgroup_task_spec.component_ref.name = subgroup_component_name

            if isinstance(subgroup, dsl.OpsGroup) and subgroup.type == 'graph':
                raise NotImplementedError(
                    'dsl.graph_component is not yet supported in KFP v2 compiler.'
                )

            if isinstance(subgroup, dsl.ContainerOp):
                if hasattr(subgroup, 'importer_spec'):
                    importer_comp_name = subgroup.task_spec.component_ref.name
                    importer_exec_label = subgroup.component_spec.executor_label
                    group_component_spec.dag.tasks[subgroup.name].CopyFrom(
                        subgroup.task_spec)
                    pipeline_spec.components[importer_comp_name].CopyFrom(
                        subgroup.component_spec)
                    deployment_config.executors[
                        importer_exec_label].importer.CopyFrom(
                            subgroup.importer_spec)

                # Task level caching option.
                subgroup.task_spec.caching_options.enable_cache = subgroup.enable_caching

            subgroup_inputs = inputs.get(subgroup.name, [])
            subgroup_params = [param for param, _ in subgroup_inputs]

            tasks_in_current_dag = [
                dsl_utils.sanitize_task_name(subgroup.name)
                for subgroup in subgroups
            ]

            input_parameters_in_current_dag = [
                input_name for input_name in
                group_component_spec.input_definitions.parameters
            ]
            input_artifacts_in_current_dag = [
                input_name for input_name in
                group_component_spec.input_definitions.artifacts
            ]

            is_parent_component_root = group_component_spec == pipeline_spec.root

            if isinstance(subgroup, dsl.ContainerOp):
                dsl_component_spec.update_task_inputs_spec(
                    subgroup_task_spec,
                    group_component_spec.input_definitions,
                    subgroup_params,
                    tasks_in_current_dag,
                    input_parameters_in_current_dag,
                    input_artifacts_in_current_dag,
                )

            if isinstance(subgroup, dsl.ParallelFor):
                if subgroup.parallelism is not None:
                    warnings.warn(
                        'Setting parallelism in ParallelFor is not supported yet.'
                        'The setting is ignored.')

                # "Punch the hole", adding additional inputs (other than loop arguments
                # which will be handled separately) needed by its subgroup or tasks.
                loop_subgroup_params = []
                for param in subgroup_params:
                    if isinstance(param, (_for_loop.LoopArguments,
                                          _for_loop.LoopArgumentVariable)):
                        continue
                    loop_subgroup_params.append(param)

                if subgroup.items_is_pipeline_param:
                    # This loop_args is a 'withParam' rather than a 'withItems'.
                    # i.e., rather than a static list, it is either the output of
                    # another task or an input as global pipeline parameters.
                    loop_subgroup_params.append(
                        subgroup.loop_args.items_or_pipeline_param)

                dsl_component_spec.build_component_inputs_spec(
                    component_spec=subgroup_component_spec,
                    pipeline_params=loop_subgroup_params,
                    is_root_component=False,
                )
                dsl_component_spec.build_task_inputs_spec(
                    subgroup_task_spec,
                    loop_subgroup_params,
                    tasks_in_current_dag,
                    is_parent_component_root,
                )

                if subgroup.items_is_pipeline_param:
                    input_parameter_name = (
                        dsl_component_spec
                        .additional_input_name_for_pipelineparam(
                            subgroup.loop_args.items_or_pipeline_param))
                    loop_arguments_item = '{}-{}'.format(
                        input_parameter_name,
                        _for_loop.LoopArguments.LOOP_ITEM_NAME_BASE)

                    subgroup_component_spec.input_definitions.parameters[
                        loop_arguments_item].type = pipeline_spec_pb2.PrimitiveType.STRING
                    subgroup_task_spec.parameter_iterator.items.input_parameter = (
                        input_parameter_name)
                    subgroup_task_spec.parameter_iterator.item_input = (
                        loop_arguments_item)

                    # If the loop arguments itself is a loop arguments variable, handle
                    # the subvar name.
                    loop_args_name, subvar_name = (
                        dsl_component_spec._exclude_loop_arguments_variables(
                            subgroup.loop_args.items_or_pipeline_param))
                    if subvar_name:
                        subgroup_task_spec.inputs.parameters[
                            input_parameter_name].parameter_expression_selector = (
                                'parseJson(string_value)["{}"]'.format(
                                    subvar_name))
                        subgroup_task_spec.inputs.parameters[
                            input_parameter_name].component_input_parameter = (
                                dsl_component_spec
                                .additional_input_name_for_pipelineparam(
                                    loop_args_name))

                else:
                    input_parameter_name = (
                        dsl_component_spec
                        .additional_input_name_for_pipelineparam(
                            subgroup.loop_args.full_name))
                    raw_values = subgroup.loop_args.to_list_for_task_yaml()

                    subgroup_component_spec.input_definitions.parameters[
                        input_parameter_name].type = pipeline_spec_pb2.PrimitiveType.STRING
                    subgroup_task_spec.parameter_iterator.items.raw = json.dumps(
                        raw_values, sort_keys=True)
                    subgroup_task_spec.parameter_iterator.item_input = (
                        input_parameter_name)

            if isinstance(subgroup,
                          dsl.OpsGroup) and subgroup.type == 'condition':

                # "punch the hole", adding inputs needed by its subgroup or tasks.
                dsl_component_spec.build_component_inputs_spec(
                    component_spec=subgroup_component_spec,
                    pipeline_params=subgroup_params,
                    is_root_component=False,
                )
                dsl_component_spec.build_task_inputs_spec(
                    subgroup_task_spec,
                    subgroup_params,
                    tasks_in_current_dag,
                    is_parent_component_root,
                )

                condition = subgroup.condition
                operand_values = []

                operand1_value, operand2_value = self._resolve_condition_operands(
                    condition.operand1, condition.operand2)

                condition_string = '{} {} {}'.format(operand1_value,
                                                     condition.operator,
                                                     operand2_value)

                subgroup_task_spec.trigger_policy.CopyFrom(
                    pipeline_spec_pb2.PipelineTaskSpec.TriggerPolicy(
                        condition=condition_string))

            if isinstance(subgroup,
                          dsl.OpsGroup) and subgroup.type == 'exit_handler':

                # "punch the hole", adding inputs needed by its subgroup or tasks.
                dsl_component_spec.build_component_inputs_spec(
                    component_spec=subgroup_component_spec,
                    pipeline_params=subgroup_params,
                    is_root_component=False,
                )
                dsl_component_spec.build_task_inputs_spec(
                    subgroup_task_spec,
                    subgroup_params,
                    tasks_in_current_dag,
                    is_parent_component_root,
                )

            # Generate dependencies section for this task.
            if dependencies.get(subgroup.name, None):
                group_dependencies = list(dependencies[subgroup.name])
                group_dependencies.sort()
                subgroup_task_spec.dependent_tasks.extend([
                    dsl_utils.sanitize_task_name(dep)
                    for dep in group_dependencies
                ])

            # Add component spec if not exists
            if subgroup_component_name not in pipeline_spec.components:
                pipeline_spec.components[subgroup_component_name].CopyFrom(
                    subgroup_component_spec)

            # Add task spec
            group_component_spec.dag.tasks[subgroup.name].CopyFrom(
                subgroup_task_spec)

            # Add AIPlatformCustomJobSpec, if applicable.
            custom_job_spec = getattr(subgroup, 'custom_job_spec', None)
            if custom_job_spec:
                executor_label = subgroup_component_spec.executor_label
                if executor_label not in deployment_config.executors:
                    deployment_config.executors[
                        executor_label].custom_job.custom_job.update(
                            custom_job_spec)

            # Add executor spec, if applicable.
            container_spec = getattr(subgroup, 'container_spec', None)
            # Ignore contaienr_spec if custom_job_spec exists.
            if container_spec and not custom_job_spec:
                if compiler_utils.is_v2_component(subgroup):
                    compiler_utils.refactor_v2_container_spec(container_spec)
                executor_label = subgroup_component_spec.executor_label

                if executor_label not in deployment_config.executors:
                    deployment_config.executors[
                        executor_label].container.CopyFrom(container_spec)

        pipeline_spec.deployment_spec.update(
            json_format.MessageToDict(deployment_config))

        # Surface metrics outputs to the top.
        self._populate_metrics_in_dag_outputs(
            group.ops,
            op_to_parent_groups,
            pipeline_spec,
        )

    def _create_pipeline_spec(
        self,
        args: List[dsl.PipelineParam],
        pipeline: dsl.Pipeline,
    ) -> pipeline_spec_pb2.PipelineSpec:
        """Creates the pipeline spec object.

        Args:
          args: The list of pipeline arguments.
          pipeline: The instantiated pipeline object.

        Returns:
          A PipelineSpec proto representing the compiled pipeline.

        Raises:
          NotImplementedError if the argument is of unsupported types.
        """
        compiler_utils.validate_pipeline_name(pipeline.name)

        deployment_config = pipeline_spec_pb2.PipelineDeploymentConfig()
        pipeline_spec = pipeline_spec_pb2.PipelineSpec()

        pipeline_spec.pipeline_info.name = pipeline.name
        pipeline_spec.sdk_version = 'kfp-{}'.format(kfp.__version__)
        # Schema version 2.0.0 is required for kfp-pipeline-spec>0.1.3.1
        pipeline_spec.schema_version = '2.0.0'

        dsl_component_spec.build_component_inputs_spec(
            component_spec=pipeline_spec.root,
            pipeline_params=args,
            is_root_component=True)

        root_group = pipeline.groups[0]
        opsgroups = self._get_groups(root_group)
        op_name_to_parent_groups = self._get_groups_for_ops(root_group)
        opgroup_name_to_parent_groups = self._get_groups_for_opsgroups(
            root_group)

        condition_params = self._get_condition_params_for_ops(root_group)
        op_name_to_for_loop_op = self._get_for_loop_ops(root_group)
        inputs, outputs = self._get_inputs_outputs(
            pipeline,
            args,
            root_group,
            op_name_to_parent_groups,
            opgroup_name_to_parent_groups,
            condition_params,
            op_name_to_for_loop_op,
        )
        dependencies = self._get_dependencies(
            pipeline,
            root_group,
            op_name_to_parent_groups,
            opgroup_name_to_parent_groups,
            opsgroups,
            condition_params,
        )

        for opsgroup_name in opsgroups.keys():
            self._group_to_dag_spec(
                opsgroups[opsgroup_name],
                inputs,
                outputs,
                dependencies,
                pipeline_spec,
                deployment_config,
                root_group.name,
                op_name_to_parent_groups,
            )

        # Exit Handler
        if pipeline.groups[0].groups:
            first_group = pipeline.groups[0].groups[0]
            if first_group.type == 'exit_handler':
                exit_handler_op = first_group.exit_op

                # Add exit op task spec
                task_name = exit_handler_op.name
                display_name = exit_handler_op.display_name

                exit_handler_op.task_spec.task_info.name = display_name or task_name
                exit_handler_op.task_spec.dependent_tasks.extend(
                    pipeline_spec.root.dag.tasks.keys())
                exit_handler_op.task_spec.trigger_policy.strategy = (
                    pipeline_spec_pb2.PipelineTaskSpec.TriggerPolicy
                    .TriggerStrategy.ALL_UPSTREAM_TASKS_COMPLETED)
                pipeline_spec.root.dag.tasks[task_name].CopyFrom(
                    exit_handler_op.task_spec)

                # Add exit op component spec if it does not exist.
                component_name = exit_handler_op.task_spec.component_ref.name
                if component_name not in pipeline_spec.components:
                    pipeline_spec.components[component_name].CopyFrom(
                        exit_handler_op.component_spec)

                # Add exit op executor spec if it does not exist.
                executor_label = exit_handler_op.component_spec.executor_label
                if executor_label not in deployment_config.executors:
                    deployment_config.executors[
                        executor_label].container.CopyFrom(
                            exit_handler_op.container_spec)
                    pipeline_spec.deployment_spec.update(
                        json_format.MessageToDict(deployment_config))

        return pipeline_spec

    def _validate_exit_handler(self, pipeline):
        """Makes sure there is only one global exit handler.

        This is temporary to be compatible with KFP v1.
        """

        def _validate_exit_handler_helper(group, exiting_op_names,
                                          handler_exists):
            if group.type == 'exit_handler':
                if handler_exists or len(exiting_op_names) > 1:
                    raise ValueError(
                        'Only one global exit_handler is allowed and all ops need to be included.'
                    )
                handler_exists = True

            if group.ops:
                exiting_op_names.extend([x.name for x in group.ops])

            for g in group.groups:
                _validate_exit_handler_helper(g, exiting_op_names,
                                              handler_exists)

        return _validate_exit_handler_helper(pipeline.groups[0], [], False)

    # TODO: Sanitizing beforehand, so that we don't need to sanitize here.
    def _sanitize_and_inject_artifact(self, pipeline: dsl.Pipeline) -> None:
        """Sanitize operator/param names and inject pipeline artifact
        location."""

        # Sanitize operator names and param names
        sanitized_ops = {}

        for op in pipeline.ops.values():
            sanitized_name = sanitize_k8s_name(op.name)
            op.name = sanitized_name
            for param in op.outputs.values():
                param.name = sanitize_k8s_name(param.name, True)
                if param.op_name:
                    param.op_name = sanitize_k8s_name(param.op_name)
            if op.output is not None and not isinstance(
                    op.output, dsl._container_op._MultipleOutputsError):
                op.output.name = sanitize_k8s_name(op.output.name, True)
                op.output.op_name = sanitize_k8s_name(op.output.op_name)
            if op.dependent_names:
                op.dependent_names = [
                    sanitize_k8s_name(name) for name in op.dependent_names
                ]
            if isinstance(op, dsl.ContainerOp) and op.file_outputs is not None:
                sanitized_file_outputs = {}
                for key in op.file_outputs.keys():
                    sanitized_file_outputs[sanitize_k8s_name(
                        key, True)] = op.file_outputs[key]
                op.file_outputs = sanitized_file_outputs
            elif isinstance(
                    op, dsl.ResourceOp) and op.attribute_outputs is not None:
                sanitized_attribute_outputs = {}
                for key in op.attribute_outputs.keys():
                    sanitized_attribute_outputs[sanitize_k8s_name(key, True)] = \
                      op.attribute_outputs[key]
                op.attribute_outputs = sanitized_attribute_outputs
            if isinstance(op, dsl.ContainerOp):
                if op.input_artifact_paths:
                    op.input_artifact_paths = {
                        sanitize_k8s_name(key, True): value
                        for key, value in op.input_artifact_paths.items()
                    }
                if op.artifact_arguments:
                    op.artifact_arguments = {
                        sanitize_k8s_name(key, True): value
                        for key, value in op.artifact_arguments.items()
                    }
            sanitized_ops[sanitized_name] = op
        pipeline.ops = sanitized_ops

    def _create_pipeline_v2(
        self,
        pipeline_func: Callable[..., Any],
        pipeline_name: Optional[str] = None,
        pipeline_parameters_override: Optional[Mapping[str, Any]] = None,
    ) -> pipeline_spec_pb2.PipelineJob:
        """Creates a pipeline instance and constructs the pipeline spec from
        it.

        Args:
          pipeline_func: Pipeline function with @dsl.pipeline decorator.
          pipeline_name: The name of the pipeline. Optional.
          pipeline_parameters_override: The mapping from parameter names to values.
            Optional.

        Returns:
          A PipelineJob proto representing the compiled pipeline.
        """

        # Create the arg list with no default values and call pipeline function.
        # Assign type information to the PipelineParam
        pipeline_meta = component_factory.extract_component_interface(
            pipeline_func)
        pipeline_name = pipeline_name or pipeline_meta.name

        pipeline_root = getattr(pipeline_func, 'pipeline_root', None)

        args_list = []
        signature = inspect.signature(pipeline_func)
        for arg_name in signature.parameters:
            arg_type = None
            for pipeline_input in pipeline_meta.inputs or []:
                if arg_name == pipeline_input.name:
                    arg_type = pipeline_input.type
                    break
            if not type_utils.is_parameter_type(arg_type):
                raise TypeError(
                    'The pipeline argument "{arg_name}" is viewed as an artifact'
                    ' due to its type "{arg_type}". And we currently do not '
                    'support passing artifacts as pipeline inputs. Consider type'
                    ' annotating the argument with a primitive type, such as '
                    '"str", "int", "float", "bool", "dict", and "list".'.format(
                        arg_name=arg_name, arg_type=arg_type))
            args_list.append(
                dsl.PipelineParam(
                    sanitize_k8s_name(arg_name, True), param_type=arg_type))

        with dsl.Pipeline(pipeline_name) as dsl_pipeline:
            pipeline_func(*args_list)

        if not dsl_pipeline.ops:
            raise ValueError('Task is missing from pipeline.')

        self._validate_exit_handler(dsl_pipeline)
        self._sanitize_and_inject_artifact(dsl_pipeline)

        # Fill in the default values.
        args_list_with_defaults = []
        if pipeline_meta.inputs:
            args_list_with_defaults = [
                dsl.PipelineParam(
                    sanitize_k8s_name(input_spec.name, True),
                    param_type=input_spec.type,
                    value=input_spec.default)
                for input_spec in pipeline_meta.inputs
            ]

        # Making the pipeline group name unique to prevent name clashes with templates
        pipeline_group = dsl_pipeline.groups[0]
        temp_pipeline_group_name = uuid.uuid4().hex
        pipeline_group.name = temp_pipeline_group_name

        pipeline_spec = self._create_pipeline_spec(
            args_list_with_defaults,
            dsl_pipeline,
        )

        pipeline_parameters = {
            param.name: param for param in args_list_with_defaults
        }
        # Update pipeline parameters override if there were any.
        pipeline_parameters_override = pipeline_parameters_override or {}
        for k, v in pipeline_parameters_override.items():
            if k not in pipeline_parameters:
                raise ValueError(
                    'Pipeline parameter {} does not match any known '
                    'pipeline argument.'.format(k))
            pipeline_parameters[k].value = v

        runtime_config = compiler_utils.build_runtime_config_spec(
            output_directory=pipeline_root,
            pipeline_parameters=pipeline_parameters)
        pipeline_job = pipeline_spec_pb2.PipelineJob(
            runtime_config=runtime_config)
        pipeline_job.pipeline_spec.update(
            json_format.MessageToDict(pipeline_spec))

        return pipeline_job

    def compile(self,
                pipeline_func: Callable[..., Any],
                package_path: str,
                pipeline_name: Optional[str] = None,
                pipeline_parameters: Optional[Mapping[str, Any]] = None,
                type_check: bool = True) -> None:
        """Compile the given pipeline function into pipeline job json.

        Args:
          pipeline_func: Pipeline function with @dsl.pipeline decorator.
          package_path: The output pipeline job .json file path. for example,
            "~/pipeline_job.json"
          pipeline_name: The name of the pipeline. Optional.
          pipeline_parameters: The mapping from parameter names to values. Optional.
          type_check: Whether to enable the type check or not, default: True.
        """
        type_check_old_value = kfp.TYPE_CHECK
        compiling_for_v2_old_value = kfp.COMPILING_FOR_V2
        try:
            kfp.TYPE_CHECK = type_check
            kfp.COMPILING_FOR_V2 = True
            pipeline_job = self._create_pipeline_v2(
                pipeline_func=pipeline_func,
                pipeline_name=pipeline_name,
                pipeline_parameters_override=pipeline_parameters)
            self._write_pipeline(pipeline_job, package_path)
        finally:
            kfp.TYPE_CHECK = type_check_old_value
            kfp.COMPILING_FOR_V2 = compiling_for_v2_old_value

    def _write_pipeline(self, pipeline_job: pipeline_spec_pb2.PipelineJob,
                        output_path: str) -> None:
        """Dump pipeline spec into json file.

        Args:
          pipeline_job: IR pipeline job spec.
          ouput_path: The file path to be written.

        Raises:
          ValueError: if the specified output path doesn't end with the acceptable
          extentions.
        """
        json_text = json_format.MessageToJson(pipeline_job, sort_keys=True)

        if output_path.endswith('.json'):
            with open(output_path, 'w') as json_file:
                json_file.write(json_text)
        else:
            raise ValueError(
                'The output path {} should ends with ".json".'.format(
                    output_path))



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cloud function that proxies HTTP request after replacing placeholders."""

import datetime
import json
import logging
import re
import traceback
from typing import Optional, Tuple

from google import auth
from google.auth.transport import urllib3


def _process_request(request):
    """Wrapper of _process_request_impl with error handling."""
    try:
        return _process_request_impl(request)
    except Exception as err:
        traceback.print_exc()
        raise err


def _preprocess_request_body(
    request_body: bytes,
    time: datetime.datetime,
) -> Tuple[str, str, Optional[bytes]]:
    """Augments the request body before sending it to CAIP Pipelines API.

    Replaces placeholders, generates unique name, removes `_discovery_url`.
    Args:
      request_body: Request body
      time: The scheduled invocation time.

    Returns:
      Tuple of (url, method, resolved_request_body).
    """
    request_str = request_body.decode('utf-8')

    # Replacing placeholders like: {{$.scheduledTime.strftime('%Y-%m-%d')}}
    request_str = re.sub(r"{{\$.scheduledTime.strftime\('([^']*)'\)}}",
                         lambda m: time.strftime(m.group(1)), request_str)

    request_json = json.loads(request_str)

    url = str(request_json['_url'])
    del request_json['_url']

    method = str(request_json['_method'])
    del request_json['_method']

    resolved_request_body = None
    if request_json:
        resolved_request_str = json.dumps(request_json)
        resolved_request_body = resolved_request_str.encode('utf-8')
    return (url, method, resolved_request_body)


def _process_request_impl(request):
    """Processes the incoming HTTP request.

    Args:
      request (flask.Request): HTTP request object.

    Returns:
      The response text or any set of values that can be turned into a Response
      object using `make_response
      <http://flask.pocoo.org/docs/1.0/api/#flask.Flask.make_response>`.
    """
    time = datetime.datetime.now()

    logging.debug('request.headers=%s', request.headers)
    logging.debug('Original request body: %s', request.data)
    (url, method, resolved_request_body) = _preprocess_request_body(
        request_body=request.data,
        time=time,
    )
    logging.debug('url=%s', url)
    logging.debug('method=%s', method)
    logging.debug('Resolved request body: %s', resolved_request_body)

    credentials, _ = auth.default()
    authorized_http = urllib3.AuthorizedHttp(credentials=credentials)
    response = authorized_http.request(
        url=url,
        method=method,
        body=resolved_request_body,
    )
    data_str = response.data.decode('utf-8')
    if response.status != 200:
        print(f'response.status={response.status}')
        print(f'response.data={data_str}')
    return (response.data, response.status)



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kfp.v2.components.component_decorator import component

from kfp.v2.components.importer_node import importer

from kfp.v2.components.types.artifact_types import (
    Artifact,
    ClassificationMetrics,
    Dataset,
    HTML,
    Markdown,
    Metrics,
    Model,
    SlicedClassificationMetrics,
)

from kfp.v2.components.types.type_annotations import (
    Input,
    Output,
    InputPath,
    OutputPath,
)

from kfp.dsl import (
    graph_component,
    pipeline,
    Condition,
    ContainerOp,
    ExitHandler,
    ParallelFor,
)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Callable, Optional, List

from kfp.v2.components import component_factory


def component(func: Optional[Callable] = None,
              *,
              base_image: Optional[str] = None,
              packages_to_install: List[str] = None,
              output_component_file: Optional[str] = None,
              install_kfp_package: bool = True,
              kfp_package_path: Optional[str] = None):
    """Decorator for Python-function based components in KFP v2.

    A lightweight component is a self-contained Python function that includes
    all necessary imports and dependencies.

    Example usage:

    from kfp.v2 import dsl
    @dsl.component
    def my_function_one(input: str, output: Output[Model]):
      ...

    @dsl.component(
      base_image='python:3.9',
      output_component_file='my_function.yaml'
    )
    def my_function_two(input: Input[Mode])):
      ...

    @dsl.pipeline(pipeline_root='...',
                  name='my-pipeline')
    def pipeline():
      my_function_one_task = my_function_one(input=...)
      my_function_two_task = my_function_two(input=my_function_one_task.outputs..

    Args:
        func: The python function to create a component from. The function
            should have type annotations for all its arguments, indicating how
            it is intended to be used (e.g. as an input/output Artifact object,
            a plain parameter, or a path to a file).
        base_image: The image to use when executing |func|. It should
            contain a default Python interpreter that is compatible with KFP.
        packages_to_install: A list of optional packages to install before
            executing |func|.
        install_kfp_package: Specifies if we should add a KFP Python package to
            |packages_to_install|. Lightweight Python functions always require
            an installation of KFP in |base_image| to work. If you specify
            a |base_image| that already contains KFP, you can set this to False.
        kfp_package_path: Specifies the location from which to install KFP. By
            default, this will try to install from PyPi using the same version
            as that used when this component was created. KFP developers can
            choose to override this to point to a Github pull request or
            other pip-compatible location when testing changes to lightweight
            Python functions.

    Returns:
        A component task factory that can be used in pipeline definitions.
    """
    if func is None:
        return functools.partial(
            component,
            base_image=base_image,
            packages_to_install=packages_to_install,
            output_component_file=output_component_file,
            install_kfp_package=install_kfp_package,
            kfp_package_path=kfp_package_path)

    return component_factory.create_component_from_func(
        func,
        base_image=base_image,
        packages_to_install=packages_to_install,
        output_component_file=output_component_file,
        install_kfp_package=install_kfp_package,
        kfp_package_path=kfp_package_path)



# Copyright 2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import json

from kfp.cli.output import print_output, OutputFormat


@click.group()
def pipeline():
    """Manage pipeline resources."""
    pass


@pipeline.command()
@click.option("-p", "--pipeline-name", help="Name of the pipeline.")
@click.option("-d", "--description", help="Description for the pipeline.")
@click.argument("package-file")
@click.pass_context
def upload(ctx, pipeline_name, package_file, description=None):
    """Upload a KFP pipeline."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]
    if not pipeline_name:
        pipeline_name = package_file.split(".")[0]

    pipeline = client.upload_pipeline(package_file, pipeline_name, description)
    _display_pipeline(pipeline, output_format)


@pipeline.command()
@click.option("-p", "--pipeline-id", help="ID of the pipeline", required=False)
@click.option("-n", "--pipeline-name", help="Name of pipeline", required=False)
@click.option(
    "-v",
    "--pipeline-version",
    help="Name of the pipeline version",
    required=True)
@click.argument("package-file")
@click.pass_context
def upload_version(ctx,
                   package_file,
                   pipeline_version,
                   pipeline_id=None,
                   pipeline_name=None):
    """Upload a version of the KFP pipeline."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]
    if bool(pipeline_id) == bool(pipeline_name):
        raise ValueError("Need to supply 'pipeline-name' or 'pipeline-id'")
    if pipeline_name is not None:
        pipeline_id = client.get_pipeline_id(name=pipeline_name)
        if pipeline_id is None:
            raise ValueError("Can't find a pipeline with name: %s" %
                             pipeline_name)
    version = client.pipeline_uploads.upload_pipeline_version(
        package_file, name=pipeline_version, pipelineid=pipeline_id)
    _display_pipeline_version(version, output_format)


@pipeline.command()
@click.option(
    "-m", "--max-size", default=100, help="Max size of the listed pipelines.")
@click.pass_context
def list(ctx, max_size):
    """List uploaded KFP pipelines."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]

    response = client.list_pipelines(
        page_size=max_size, sort_by="created_at desc")
    if response.pipelines:
        _print_pipelines(response.pipelines, output_format)
    else:
        if output_format == OutputFormat.json.name:
            msg = json.dumps([])
        else:
            msg = "No pipelines found"
        click.echo(msg)


@pipeline.command()
@click.argument("pipeline-id")
@click.option(
    "-m", "--max-size", default=10, help="Max size of the listed pipelines.")
@click.pass_context
def list_versions(ctx, pipeline_id, max_size):
    """List versions of an uploaded KFP pipeline."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]

    response = client.list_pipeline_versions(
        pipeline_id=pipeline_id, page_size=max_size, sort_by="created_at desc")
    if response.versions:
        _print_pipeline_versions(response.versions, output_format)
    else:
        if output_format == OutputFormat.json.name:
            msg = json.dumps([])
        else:
            msg = "No pipeline or version found"
        click.echo(msg)


@pipeline.command()
@click.argument("pipeline-id")
@click.pass_context
def get(ctx, pipeline_id):
    """Get detailed information about an uploaded KFP pipeline."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]

    pipeline = client.get_pipeline(pipeline_id)
    _display_pipeline(pipeline, output_format)


@pipeline.command()
@click.argument("pipeline-id")
@click.pass_context
def delete(ctx, pipeline_id):
    """Delete an uploaded KFP pipeline."""
    client = ctx.obj["client"]

    client.delete_pipeline(pipeline_id)
    click.echo("{} is deleted".format(pipeline_id))


def _print_pipelines(pipelines, output_format):
    headers = ["Pipeline ID", "Name", "Uploaded at"]
    data = [[pipeline.id, pipeline.name,
             pipeline.created_at.isoformat()] for pipeline in pipelines]
    print_output(data, headers, output_format, table_format="grid")


def _print_pipeline_versions(versions, output_format):
    headers = ["Version ID", "Version name", "Uploaded at"]
    data = [[version.id, version.name,
             version.created_at.isoformat()] for version in versions]
    print_output(data, headers, output_format, table_format="grid")


def _display_pipeline(pipeline, output_format):
    # Pipeline information
    table = [
        ["ID", pipeline.id],
        ["Name", pipeline.name],
        ["Description", pipeline.description],
        ["Uploaded at", pipeline.created_at.isoformat()],
    ]

    # Pipeline parameter details
    headers = ["Parameter Name", "Default Value"]
    data = []
    if pipeline.parameters is not None:
        data = [[param.name, param.value] for param in pipeline.parameters]

    if output_format == OutputFormat.table.name:
        print_output([], ["Pipeline Details"], output_format)
        print_output(table, [], output_format, table_format="plain")
        print_output(data, headers, output_format, table_format="grid")
    elif output_format == OutputFormat.json.name:
        output = dict()
        output["Pipeline Details"] = dict(table)
        params = []
        for item in data:
            params.append(dict(zip(headers, item)))
        output["Pipeline Parameters"] = params
        print_output(output, [], output_format)


def _display_pipeline_version(version, output_format):
    pipeline_id = version.resource_references[0].key.id
    table = [
        ["Pipeline ID", pipeline_id],
        ["Version Name", version.name],
        ["Uploaded at", version.created_at.isoformat()],
    ]

    if output_format == OutputFormat.table.name:
        print_output([], ["Pipeline Version Details"], output_format)
        print_output(table, [], output_format, table_format="plain")
    elif output_format == OutputFormat.json.name:
        print_output(dict(table), [], output_format)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for supporting Google Cloud AI Platform Custom Job."""

import copy
from typing import Any, List, Mapping, Optional

from kfp import dsl
from kfp.dsl import dsl_utils

_DEFAULT_CUSTOM_JOB_MACHINE_TYPE = 'n1-standard-4'


def run_as_aiplatform_custom_job(
    op: dsl.ContainerOp,
    display_name: Optional[str] = None,
    replica_count: Optional[int] = None,
    machine_type: Optional[str] = None,
    accelerator_type: Optional[str] = None,
    accelerator_count: Optional[int] = None,
    boot_disk_type: Optional[str] = None,
    boot_disk_size_gb: Optional[int] = None,
    timeout: Optional[str] = None,
    restart_job_on_worker_restart: Optional[bool] = None,
    service_account: Optional[str] = None,
    network: Optional[str] = None,
    output_uri_prefix: Optional[str] = None,
    worker_pool_specs: Optional[List[Mapping[str, Any]]] = None,
) -> None:
    """Run a pipeline task using AI Platform (Unified) custom training job.

    For detailed doc of the service, please refer to
    https://cloud.google.com/ai-platform-unified/docs/training/create-custom-job

    Args:
      op: The task (ContainerOp) object to run as aiplatform custom job.
      display_name: Optional. The name of the custom job.
      replica_count: Optional. The number of replicas to be split between master
        workerPoolSpec and worker workerPoolSpec. (master always has 1 replica).
      machine_type: Optional. The type of the machine to run the custom job. The
        default value is "n1-standard-4".
      accelerator_type: Optional. The type of accelerator(s) that may be attached
        to the machine as per acceleratorCount. Optional.
      accelerator_count: Optional. The number of accelerators to attach to the
        machine.
      boot_disk_type: Optional. Type of the boot disk (default is "pd-ssd"). Valid
        values: "pd-ssd" (Persistent Disk Solid State Drive) or "pd-standard"
          (Persistent Disk Hard Disk Drive).
      boot_disk_size_gb: Optional. Size in GB of the boot disk (default is 100GB).
      timeout: Optional. The maximum job running time. The default is 7 days. A
        duration in seconds with up to nine fractional digits, terminated by 's'.
        Example: "3.5s"
      restart_job_on_worker_restart: Optional. Restarts the entire CustomJob if a
        worker gets restarted. This feature can be used by distributed training
        jobs that are not resilient to workers leaving and joining a job.
      service_account: Optional. Specifies the service account for workload run-as
        account.
      network: Optional. The full name of the Compute Engine network to which the
        job should be peered. For example, projects/12345/global/networks/myVPC.
      output_uri_prefix: Optional. Google Cloud Storage URI to output directory.
      additional_worker_pool_specs: Optional. Additional workerPoolSpecs for
        distributed training. For details, please see:
        https://cloud.google.com/ai-platform-unified/docs/training/distributed-training
    """
    job_spec = {}

    if worker_pool_specs is not None:
        worker_pool_specs = copy.deepcopy(worker_pool_specs)

        def _is_output_parameter(output_key: str) -> bool:
            return output_key in (
                op.component_spec.output_definitions.parameters.keys())

        for worker_pool_spec in worker_pool_specs:
            if 'containerSpec' in worker_pool_spec:
                container_spec = worker_pool_spec['containerSpec']
                if 'command' in container_spec:
                    dsl_utils.resolve_cmd_lines(container_spec['command'],
                                                _is_output_parameter)
                if 'args' in container_spec:
                    dsl_utils.resolve_cmd_lines(container_spec['args'],
                                                _is_output_parameter)

            elif 'pythonPackageSpec' in worker_pool_spec:
                # For custom Python training, resolve placeholders in args only.
                python_spec = worker_pool_spec['pythonPackageSpec']
                if 'args' in python_spec:
                    dsl_utils.resolve_cmd_lines(python_spec['args'],
                                                _is_output_parameter)

            else:
                raise ValueError(
                    'Expect either "containerSpec" or "pythonPackageSpec" in each '
                    'workerPoolSpec. Got: {}'.format(custom_job_spec))

        job_spec['workerPoolSpecs'] = worker_pool_specs

    else:
        worker_pool_spec = {
            'machineSpec': {
                'machineType': machine_type or _DEFAULT_CUSTOM_JOB_MACHINE_TYPE
            },
            'replicaCount': '1',
            'containerSpec': {
                'imageUri': op.container.image,
            }
        }
        if op.container.command:
            worker_pool_spec['containerSpec']['command'] = op.container.command
        if op.container.args:
            worker_pool_spec['containerSpec']['args'] = op.container.args
        if accelerator_type is not None:
            worker_pool_spec['machineSpec'][
                'acceleratorType'] = accelerator_type
        if accelerator_count is not None:
            worker_pool_spec['machineSpec'][
                'acceleratorCount'] = accelerator_count
        if boot_disk_type is not None:
            if 'diskSpec' not in worker_pool_spec:
                worker_pool_spec['diskSpec'] = {}
            worker_pool_spec['diskSpec']['bootDiskType'] = boot_disk_type
        if boot_disk_size_gb is not None:
            if 'diskSpec' not in worker_pool_spec:
                worker_pool_spec['diskSpec'] = {}
            worker_pool_spec['diskSpec']['bootDiskSizeGb'] = boot_disk_size_gb

        job_spec['workerPoolSpecs'] = [worker_pool_spec]
        if replica_count is not None and replica_count > 1:
            additional_worker_pool_spec = copy.deepcopy(worker_pool_spec)
            additional_worker_pool_spec['replicaCount'] = str(replica_count - 1)
            job_spec['workerPoolSpecs'].append(additional_worker_pool_spec)

    if timeout is not None:
        if 'scheduling' not in job_spec:
            job_spec['scheduling'] = {}
        job_spec['scheduling']['timeout'] = timeout
    if restart_job_on_worker_restart is not None:
        if 'scheduling' not in job_spec:
            job_spec['scheduling'] = {}
        job_spec['scheduling'][
            'restartJobOnWorkerRestart'] = restart_job_on_worker_restart
    if service_account is not None:
        job_spec['serviceAccount'] = service_account
    if network is not None:
        job_spec['network'] = network
    if output_uri_prefix is not None:
        job_spec['baseOutputDirectory'] = {'outputUriPrefix': output_uri_prefix}

    op.custom_job_spec = {
        'displayName': display_name or op.name,
        'jobSpec': job_spec
    }



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.v2.google.client.client_utils."""

import json
import unittest
from unittest import mock

from google.cloud import storage

from kfp.v2.google.client import client_utils


class ClientUtilsTest(unittest.TestCase):

    @mock.patch.object(storage, 'Client', autospec=True)
    @mock.patch.object(storage.Blob, 'download_as_bytes', autospec=True)
    def test_load_json_from_gs_uri(self, mock_download_as_bytes,
                                   unused_storage_client):
        mock_download_as_bytes.return_value = b'{"key":"value"}'
        self.assertEqual({'key': 'value'},
                         client_utils.load_json('gs://bucket/path/to/blob'))

    @mock.patch('builtins.open', mock.mock_open(read_data='{"key":"value"}'))
    def test_load_json_from_local_file(self):
        self.assertEqual({'key': 'value'},
                         client_utils.load_json('/path/to/file'))

    @mock.patch.object(storage, 'Client', autospec=True)
    def test_load_json_from_gs_uri_with_non_gs_uri_should_fail(
            self, unused_storage_client):
        with self.assertRaisesRegex(ValueError, 'URI scheme must be gs'):
            client_utils._load_json_from_gs_uri(
                'https://storage.google.com/bucket/blob')

    @mock.patch.object(storage, 'Client', autospec=True)
    @mock.patch.object(storage.Blob, 'download_as_bytes', autospec=True)
    def test_load_json_from_gs_uri_with_invalid_json_should_fail(
            self, mock_download_as_bytes, unused_storage_client):
        mock_download_as_bytes.return_value = b'invalid-json'
        with self.assertRaises(json.decoder.JSONDecodeError):
            client_utils._load_json_from_gs_uri('gs://bucket/path/to/blob')

    @mock.patch('builtins.open', mock.mock_open(read_data='invalid-json'))
    def test_load_json_from_local_file_with_invalid_json_should_fail(self):
        with self.assertRaises(json.decoder.JSONDecodeError):
            client_utils._load_json_from_local_file('/path/to/file')


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from absl.testing import parameterized
from google.protobuf import json_format
from kfp.dsl import _pipeline_param
from kfp.pipeline_spec import pipeline_spec_pb2 as pb
from kfp.v2.components import importer_node


class ImporterNodeTest(parameterized.TestCase):

    @parameterized.parameters(
        {
            # artifact_uri is a constant value
            'input_uri':
                'gs://artifact',
            'artifact_type_schema':
                pb.ArtifactTypeSchema(schema_title='system.Dataset'),
            'expected_result': {
                'artifactUri': {
                    'constantValue': {
                        'stringValue': 'gs://artifact'
                    }
                },
                'typeSchema': {
                    'schemaTitle': 'system.Dataset'
                }
            }
        },
        {
            # artifact_uri is from PipelineParam
            'input_uri':
                _pipeline_param.PipelineParam(name='uri_to_import'),
            'artifact_type_schema':
                pb.ArtifactTypeSchema(schema_title='system.Model'),
            'expected_result': {
                'artifactUri': {
                    'runtimeParameter': 'uri'
                },
                'typeSchema': {
                    'schemaTitle': 'system.Model'
                }
            },
        })
    def test_build_importer_spec(self, input_uri, artifact_type_schema,
                                 expected_result):
        expected_importer_spec = pb.PipelineDeploymentConfig.ImporterSpec()
        json_format.ParseDict(expected_result, expected_importer_spec)
        importer_spec = importer_node._build_importer_spec(
            artifact_uri=input_uri, artifact_type_schema=artifact_type_schema)

        self.maxDiff = None
        self.assertEqual(expected_importer_spec, importer_spec)

    @parameterized.parameters(
        {
            # artifact_uri is a constant value
            'importer_name': 'importer-1',
            'input_uri': 'gs://artifact',
            'expected_result': {
                'inputs': {
                    'parameters': {
                        'uri': {
                            'runtimeValue': {
                                'constantValue': {
                                    'stringValue': 'gs://artifact'
                                }
                            }
                        }
                    }
                },
                'componentRef': {
                    'name': 'comp-importer-1'
                },
            }
        },
        {
            # artifact_uri is from PipelineParam
            'importer_name': 'importer-2',
            'input_uri': _pipeline_param.PipelineParam(name='uri_to_import'),
            'expected_result': {
                'inputs': {
                    'parameters': {
                        'uri': {
                            'componentInputParameter': 'uri_to_import'
                        }
                    }
                },
                'componentRef': {
                    'name': 'comp-importer-2'
                },
            },
        })
    def test_build_importer_task_spec(self, importer_name, input_uri,
                                      expected_result):
        expected_task_spec = pb.PipelineTaskSpec()
        json_format.ParseDict(expected_result, expected_task_spec)

        task_spec = importer_node._build_importer_task_spec(
            importer_base_name=importer_name, artifact_uri=input_uri)

        self.maxDiff = None
        self.assertEqual(expected_task_spec, task_spec)

    def test_build_importer_component_spec(self):
        expected_importer_component = {
            'inputDefinitions': {
                'parameters': {
                    'uri': {
                        'type': 'STRING'
                    }
                }
            },
            'outputDefinitions': {
                'artifacts': {
                    'artifact': {
                        'artifactType': {
                            'schemaTitle': 'system.Artifact'
                        }
                    }
                }
            },
            'executorLabel': 'exec-importer-1'
        }
        expected_importer_comp_spec = pb.ComponentSpec()
        json_format.ParseDict(expected_importer_component,
                              expected_importer_comp_spec)
        importer_comp_spec = importer_node._build_importer_component_spec(
            importer_base_name='importer-1',
            artifact_type_schema=pb.ArtifactTypeSchema(
                schema_title='system.Artifact'))

        self.maxDiff = None
        self.assertEqual(expected_importer_comp_spec, importer_comp_spec)

    def test_import_with_invalid_artifact_uri_value_should_fail(self):
        from kfp.v2.components.types.artifact_types import Dataset
        with self.assertRaisesRegex(
                ValueError,
                "Importer got unexpected artifact_uri: 123 of type: <class 'int'>."
        ):
            importer_node.importer(artifact_uri=123, artifact_class=Dataset)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.components.executor."""

import json
import os
import tempfile
import unittest
from typing import Callable, Dict, List, NamedTuple, Optional

from kfp.v2.components import executor
from kfp.v2.components.types import artifact_types
from kfp.v2.components.types.artifact_types import (Artifact, Dataset, Metrics,
                                                    Model)
from kfp.v2.components.types.type_annotations import (Input, InputPath, Output,
                                                      OutputPath)

_EXECUTOR_INPUT = """\
{
  "inputs": {
    "parameters": {
      "input_parameter": {
        "stringValue": "Hello, KFP"
      }
    },
    "artifacts": {
      "input_artifact_one_path": {
        "artifacts": [
          {
            "metadata": {},
            "name": "input_artifact_one",
            "type": {
              "schemaTitle": "system.Dataset"
            },
            "uri": "gs://some-bucket/input_artifact_one"
          }
        ]
      }
    }
  },
  "outputs": {
    "artifacts": {
      "output_artifact_one_path": {
        "artifacts": [
          {
            "metadata": {},
            "name": "output_artifact_one",
            "type": {
              "schemaTitle": "system.Model"
            },
            "uri": "gs://some-bucket/output_artifact_one"
          }
        ]
      },
      "output_artifact_two": {
        "artifacts": [
          {
            "metadata": {},
            "name": "output_artifact_two",
            "type": {
              "schemaTitle": "system.Metrics"
            },
            "uri": "gs://some-bucket/output_artifact_two"
          }
        ]
      }
    },
    "parameters": {
      "output_parameter_path": {
        "outputFile": "gs://some-bucket/some_task/nested/output_parameter"
      }
    },
    "outputFile": "%s/output_metadata.json"
  }
}
"""


class ExecutorTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self._test_dir = tempfile.mkdtemp()
        artifact_types._GCS_LOCAL_MOUNT_PREFIX = self._test_dir + '/'
        artifact_types._MINIO_LOCAL_MOUNT_PREFIX = self._test_dir + '/minio/'
        artifact_types._S3_LOCAL_MOUNT_PREFIX = self._test_dir + '/s3/'
        return super().setUp()

    def _get_executor(
            self,
            func: Callable,
            executor_input: Optional[str] = None) -> executor.Executor:
        if executor_input is None:
            executor_input = _EXECUTOR_INPUT

        executor_input_dict = json.loads(executor_input % self._test_dir)

        return executor.Executor(
            executor_input=executor_input_dict, function_to_execute=func)

    def test_input_parameter(self):

        def test_func(input_parameter: str):
            self.assertEqual(input_parameter, "Hello, KFP")

        self._get_executor(test_func).execute()

    def test_input_artifact(self):

        def test_func(input_artifact_one_path: Input[Dataset]):
            self.assertEqual(input_artifact_one_path.uri,
                             'gs://some-bucket/input_artifact_one')
            self.assertEqual(
                input_artifact_one_path.path,
                os.path.join(self._test_dir, 'some-bucket/input_artifact_one'))
            self.assertEqual(input_artifact_one_path.name, 'input_artifact_one')

        self._get_executor(test_func).execute()

    def test_output_artifact(self):

        def test_func(output_artifact_one_path: Output[Model]):
            self.assertEqual(output_artifact_one_path.uri,
                             'gs://some-bucket/output_artifact_one')

            self.assertEqual(
                output_artifact_one_path.path,
                os.path.join(self._test_dir, 'some-bucket/output_artifact_one'))
            self.assertEqual(output_artifact_one_path.name,
                             'output_artifact_one')

        self._get_executor(test_func).execute()

    def test_output_parameter(self):

        def test_func(output_parameter_path: OutputPath(str)):
            # Test that output parameters just use the passed in filename.
            self.assertEqual(
                output_parameter_path,
                'gs://some-bucket/some_task/nested/output_parameter')

            # Test writing to the path succeeds. This fails if parent directories
            # don't exist.
            with open(output_parameter_path, 'w') as f:
                f.write('Hello, World!')

        self._get_executor(test_func).execute()

    def test_input_path_artifact(self):

        def test_func(input_artifact_one_path: InputPath('Dataset')):
            self.assertEqual(
                input_artifact_one_path,
                os.path.join(self._test_dir, 'some-bucket/input_artifact_one'))

        self._get_executor(test_func).execute()

    def test_output_path_artifact(self):

        def test_func(output_artifact_one_path: OutputPath('Model')):
            self.assertEqual(
                output_artifact_one_path,
                os.path.join(self._test_dir, 'some-bucket/output_artifact_one'))

        self._get_executor(test_func).execute()

    def test_output_metadata(self):

        def test_func(output_artifact_two: Output[Metrics]):
            output_artifact_two.metadata['key_1'] = 'value_1'
            output_artifact_two.metadata['key_2'] = 2
            output_artifact_two.uri = 'new-uri'

            # log_metric works here since the schema is specified as Metrics.
            output_artifact_two.log_metric('metric', 0.9)

        self._get_executor(test_func).execute()
        with open(os.path.join(self._test_dir, 'output_metadata.json'),
                  'r') as f:
            output_metadata = json.loads(f.read())
        self.assertDictEqual(
            output_metadata, {
                'artifacts': {
                    'output_artifact_one_path': {
                        'artifacts': [{
                            'name': 'output_artifact_one',
                            'uri': 'gs://some-bucket/output_artifact_one',
                            'metadata': {}
                        }]
                    },
                    'output_artifact_two': {
                        'artifacts': [{
                            'name': 'output_artifact_two',
                            'uri': 'new-uri',
                            'metadata': {
                                'key_1': 'value_1',
                                'key_2': 2,
                                'metric': 0.9
                            }
                        }]
                    }
                }
            })

    def test_function_string_output(self):
        executor_input = """\
    {
      "inputs": {
        "parameters": {
          "first_message": {
            "stringValue": "Hello"
          },
          "second_message": {
            "stringValue": ""
          },
          "third_message": {
            "stringValue": "World"
          }
        }
      },
      "outputs": {
        "parameters": {
          "output": {
            "outputFile": "gs://some-bucket/output"
          }
        },
        "outputFile": "%s/output_metadata.json"
      }
    }
    """

        def test_func(
            first_message: str,
            second_message: str,
            third_message: str,
        ) -> str:
            return first_message + ", " + second_message + ", " + third_message

        self._get_executor(test_func, executor_input).execute()
        with open(os.path.join(self._test_dir, 'output_metadata.json'),
                  'r') as f:
            output_metadata = json.loads(f.read())
        self.assertDictEqual(output_metadata, {
            "parameters": {
                "Output": {
                    "stringValue": "Hello, , World"
                }
            },
        })

    def test_function_with_int_output(self):
        executor_input = """\
    {
      "inputs": {
        "parameters": {
          "first": {
            "intValue": 40
          },
          "second": {
            "intValue": 2
          }
        }
      },
      "outputs": {
        "parameters": {
          "output": {
            "outputFile": "gs://some-bucket/output"
          }
        },
        "outputFile": "%s/output_metadata.json"
      }
    }
    """

        def test_func(first: int, second: int) -> int:
            return first + second

        self._get_executor(test_func, executor_input).execute()
        with open(os.path.join(self._test_dir, 'output_metadata.json'),
                  'r') as f:
            output_metadata = json.loads(f.read())
        self.assertDictEqual(output_metadata, {
            "parameters": {
                "Output": {
                    "intValue": 42
                }
            },
        })

    def test_function_with_list_output(self):
        executor_input = """\
    {
      "inputs": {
        "parameters": {
          "first": {
            "intValue": 40
          },
          "second": {
            "intValue": 2
          }
        }
      },
      "outputs": {
        "parameters": {
          "output": {
            "outputFile": "gs://some-bucket/output"
          }
        },
        "outputFile": "%s/output_metadata.json"
      }
    }
    """

        def test_func(first: int, second: int) -> List:
            return [first, second]

        self._get_executor(test_func, executor_input).execute()
        with open(os.path.join(self._test_dir, 'output_metadata.json'),
                  'r') as f:
            output_metadata = json.loads(f.read())
        self.assertDictEqual(output_metadata, {
            "parameters": {
                "Output": {
                    "stringValue": "[40, 2]"
                }
            },
        })

    def test_function_with_dict_output(self):
        executor_input = """\
    {
      "inputs": {
        "parameters": {
          "first": {
            "intValue": 40
          },
          "second": {
            "intValue": 2
          }
        }
      },
      "outputs": {
        "parameters": {
          "output": {
            "outputFile": "gs://some-bucket/output"
          }
        },
        "outputFile": "%s/output_metadata.json"
      }
    }
    """

        def test_func(first: int, second: int) -> Dict:
            return {"first": first, "second": second}

        self._get_executor(test_func, executor_input).execute()
        with open(os.path.join(self._test_dir, 'output_metadata.json'),
                  'r') as f:
            output_metadata = json.loads(f.read())
        self.assertDictEqual(output_metadata, {
            "parameters": {
                "Output": {
                    "stringValue": "{\"first\": 40, \"second\": 2}"
                }
            },
        })

    def test_function_with_typed_list_output(self):
        executor_input = """\
    {
      "inputs": {
        "parameters": {
          "first": {
            "intValue": 40
          },
          "second": {
            "intValue": 2
          }
        }
      },
      "outputs": {
        "parameters": {
          "output": {
            "outputFile": "gs://some-bucket/output"
          }
        },
        "outputFile": "%s/output_metadata.json"
      }
    }
    """

        def test_func(first: int, second: int) -> List[int]:
            return [first, second]

        self._get_executor(test_func, executor_input).execute()
        with open(os.path.join(self._test_dir, 'output_metadata.json'),
                  'r') as f:
            output_metadata = json.loads(f.read())
        self.assertDictEqual(output_metadata, {
            "parameters": {
                "Output": {
                    "stringValue": "[40, 2]"
                }
            },
        })

    def test_function_with_typed_dict_output(self):
        executor_input = """\
    {
      "inputs": {
        "parameters": {
          "first": {
            "intValue": 40
          },
          "second": {
            "intValue": 2
          }
        }
      },
      "outputs": {
        "parameters": {
          "output": {
            "outputFile": "gs://some-bucket/output"
          }
        },
        "outputFile": "%s/output_metadata.json"
      }
    }
    """

        def test_func(first: int, second: int) -> Dict[str, int]:
            return {"first": first, "second": second}

        self._get_executor(test_func, executor_input).execute()
        with open(os.path.join(self._test_dir, 'output_metadata.json'),
                  'r') as f:
            output_metadata = json.loads(f.read())
        self.assertDictEqual(output_metadata, {
            "parameters": {
                "Output": {
                    "stringValue": "{\"first\": 40, \"second\": 2}"
                }
            },
        })

    def test_artifact_output(self):
        executor_input = """\
    {
      "inputs": {
        "parameters": {
          "first": {
            "stringValue": "Hello"
          },
          "second": {
            "stringValue": "World"
          }
        }
      },
      "outputs": {
        "artifacts": {
          "Output": {
            "artifacts": [
              {
                "name": "output",
                "type": {
                  "schemaTitle": "system.Artifact"
                },
                "uri": "gs://some-bucket/output"
              }
            ]
          }
        },
        "outputFile": "%s/output_metadata.json"
      }
    }
    """

        def test_func(first: str, second: str) -> Artifact:
            return first + ", " + second

        self._get_executor(test_func, executor_input).execute()
        with open(os.path.join(self._test_dir, 'output_metadata.json'),
                  'r') as f:
            output_metadata = json.loads(f.read())
        self.assertDictEqual(
            output_metadata, {
                'artifacts': {
                    'Output': {
                        'artifacts': [{
                            'metadata': {},
                            'name': 'output',
                            'uri': 'gs://some-bucket/output'
                        }]
                    }
                }
            })

        with open(os.path.join(self._test_dir, 'some-bucket/output'), 'r') as f:
            artifact_payload = f.read()
        self.assertEqual(artifact_payload, "Hello, World")

    def test_named_tuple_output(self):
        executor_input = """\
    {
      "outputs": {
        "artifacts": {
          "output_dataset": {
            "artifacts": [
              {
                "name": "output_dataset",
                "type": {
                  "schemaTitle": "system.Dataset"
                },
                "uri": "gs://some-bucket/output_dataset"
              }
            ]
          }
        },
        "parameters": {
          "output_int": {
            "outputFile": "gs://some-bucket/output_int"
          },
          "output_string": {
            "outputFile": "gs://some-bucket/output_string"
          }
        },
        "outputFile": "%s/output_metadata.json"
      }
    }
    """

        # Functions returning named tuples should work.
        def func_returning_named_tuple() -> NamedTuple('Outputs', [
            ("output_dataset", Dataset),
            ("output_int", int),
            ("output_string", str),
        ]):
            from collections import namedtuple
            output = namedtuple(
                'Outputs', ['output_dataset', 'output_int', 'output_string'])
            return output("Dataset contents", 101, "Some output string")

        # Functions returning plain tuples should work too.
        def func_returning_plain_tuple() -> NamedTuple('Outputs', [
            ("output_dataset", Dataset),
            ("output_int", int),
            ("output_string", str),
        ]):
            return ("Dataset contents", 101, "Some output string")

        for test_func in [
                func_returning_named_tuple, func_returning_plain_tuple
        ]:
            self._get_executor(test_func, executor_input).execute()
            with open(
                    os.path.join(self._test_dir, 'output_metadata.json'),
                    'r') as f:
                output_metadata = json.loads(f.read())
            self.assertDictEqual(
                output_metadata, {
                    'artifacts': {
                        'output_dataset': {
                            'artifacts': [{
                                'metadata': {},
                                'name': 'output_dataset',
                                'uri': 'gs://some-bucket/output_dataset'
                            }]
                        }
                    },
                    "parameters": {
                        "output_string": {
                            "stringValue": "Some output string"
                        },
                        "output_int": {
                            "intValue": 101
                        }
                    },
                })

            with open(
                    os.path.join(self._test_dir, 'some-bucket/output_dataset'),
                    'r') as f:
                artifact_payload = f.read()
            self.assertEqual(artifact_payload, "Dataset contents")

    def test_function_with_optional_inputs(self):
        executor_input = """\
    {
      "inputs": {
        "parameters": {
          "first_message": {
            "stringValue": "Hello"
          },
          "second_message": {
            "stringValue": "World"
          }
        }
      },
      "outputs": {
        "parameters": {
          "output": {
            "outputFile": "gs://some-bucket/output"
          }
        },
        "outputFile": "%s/output_metadata.json"
      }
    }
    """

        def test_func(
            first_message: str = 'default value',
            second_message: Optional[str] = None,
            third_message: Optional[str] = None,
        ) -> str:
            return (f'{first_message} ({type(first_message)}), '
                    f'{second_message} ({type(second_message)}), '
                    f'{third_message} ({type(third_message)}).')

        self._get_executor(test_func, executor_input).execute()
        with open(os.path.join(self._test_dir, 'output_metadata.json'),
                  'r') as f:
            output_metadata = json.loads(f.read())
        self.assertDictEqual(
            output_metadata, {
                "parameters": {
                    "Output": {
                        "stringValue": "Hello (<class 'str'>), "
                                       "World (<class 'str'>), "
                                       "None (<class 'NoneType'>)."
                    }
                },
            })


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for MLMD artifact in KFP SDK."""

from typing import Any, Dict, Optional

from absl import logging
import enum
import importlib
import jsonschema
import yaml

from google.protobuf import json_format
from google.protobuf import struct_pb2
from kfp.pipeline_spec import pipeline_spec_pb2
from kfp.dsl import serialization_utils
from kfp.dsl import artifact_utils

KFP_ARTIFACT_ONTOLOGY_MODULE = 'kfp.dsl.ontology_artifacts'
DEFAULT_ARTIFACT_SCHEMA = 'title: kfp.Artifact\ntype: object\nproperties:\n'


class Artifact(object):
    """KFP Artifact Python class.

    Artifact Python class/object mainly serves following purposes in different
    period of its lifecycle.

    1. During compile time, users can use Artifact class to annotate I/O types of
       their components.
    2. At runtime, Artifact objects provide helper function/utilities to access
       the underlying RuntimeArtifact pb message, and provide additional layers
       of validation to ensure type compatibility for fields specified in the
       instance schema.
    """

    TYPE_NAME = "kfp.Artifact"

    # Initialization flag to support setattr / getattr behavior.
    _initialized = False

    def __init__(self, instance_schema: Optional[str] = None):
        """Constructs an instance of Artifact.

        Setups up self._metadata_fields to perform type checking and
        initialize RuntimeArtifact.
        """
        if self.__class__ == Artifact:
            if not instance_schema:
                raise ValueError(
                    'The "instance_schema" argument must be set for Artifact.')
            self._instance_schema = instance_schema
        else:
            if instance_schema:
                raise ValueError(
                    'The "instance_schema" argument must not be passed for Artifact \
               subclass: {}'.format(self.__class__))

        # setup self._metadata_fields
        self.TYPE_NAME, self._metadata_fields = artifact_utils.parse_schema(
            self._instance_schema)

        # Instantiate a RuntimeArtifact pb message as the POD data structure.
        self._artifact = pipeline_spec_pb2.RuntimeArtifact()

        # Stores the metadata for the Artifact.
        self.metadata = {}

        self._artifact.type.CopyFrom(
            pipeline_spec_pb2.ArtifactTypeSchema(
                instance_schema=self._instance_schema))

        self._initialized = True

    @property
    def type_schema(self) -> str:
        """Gets the instance_schema for this Artifact object."""
        return self._instance_schema

    def __getattr__(self, name: str) -> Any:
        """Custom __getattr__ to allow access to artifact metadata."""

        if name not in self._metadata_fields:
            raise AttributeError(
                'No metadata field: {} in artifact.'.format(name))

        return self.metadata[name]

    def __setattr__(self, name: str, value: Any):
        """Custom __setattr__ to allow access to artifact metadata."""

        if not self._initialized:
            object.__setattr__(self, name, value)
            return

        metadata_fields = {}
        if self._metadata_fields:
            metadata_fields = self._metadata_fields

        if name not in self._metadata_fields:
            if (name in self.__dict__ or
                    any(name in c.__dict__ for c in self.__class__.mro())):
                # Use any provided getter / setter if available.
                object.__setattr__(self, name, value)
                return
            # In the case where we do not handle this via an explicit getter /
            # setter, we assume that the user implied an artifact attribute store,
            # and we raise an exception since such an attribute was not explicitly
            # defined in the Artifact PROPERTIES dictionary.
            raise AttributeError('Cannot set an unspecified metadata field:{} \
        on artifact. Only fields specified in instance schema can be \
        set.'.format(name))

        # Type checking to be performed during serialization.
        self.metadata[name] = value

    def _update_runtime_artifact(self):
        """Verifies metadata is well-formed and updates artifact instance."""

        artifact_utils.verify_schema_instance(self._instance_schema,
                                              self.metadata)

        if len(self.metadata) != 0:
            metadata_protobuf_struct = struct_pb2.Struct()
            metadata_protobuf_struct.update(self.metadata)
            self._artifact.metadata.CopyFrom(metadata_protobuf_struct)

    @property
    def type(self):
        return self.__class__

    @property
    def type_name(self):
        return self.TYPE_NAME

    @property
    def uri(self) -> str:
        return self._artifact.uri

    @uri.setter
    def uri(self, uri: str) -> None:
        self._artifact.uri = uri

    @property
    def name(self) -> str:
        return self._artifact.name

    @name.setter
    def name(self, name: str) -> None:
        self._artifact.name = name

    @property
    def runtime_artifact(self) -> pipeline_spec_pb2.RuntimeArtifact:
        self._update_runtime_artifact()
        return self._artifact

    @runtime_artifact.setter
    def runtime_artifact(self, artifact: pipeline_spec_pb2.RuntimeArtifact):
        self._artifact = artifact

    def serialize(self) -> str:
        """Serializes an Artifact to JSON dict format."""
        self._update_runtime_artifact()
        return json_format.MessageToJson(self._artifact, sort_keys=True)

    @classmethod
    def get_artifact_type(cls) -> str:
        """Gets the instance_schema according to the Python schema spec."""
        result_map = {'title': cls.TYPE_NAME, 'type': 'object'}

        return serialization_utils.yaml_dump(result_map)

    @classmethod
    def get_ir_type(cls) -> pipeline_spec_pb2.ArtifactTypeSchema:
        return pipeline_spec_pb2.ArtifactTypeSchema(
            instance_schema=cls.get_artifact_type())

    @classmethod
    def get_from_runtime_artifact(
            cls, artifact: pipeline_spec_pb2.RuntimeArtifact) -> Any:
        """Deserializes an Artifact object from RuntimeArtifact message."""
        instance_schema = yaml.safe_load(artifact.type.instance_schema)
        type_name = instance_schema['title'][len('kfp.'):]
        result = None
        try:
            artifact_cls = getattr(
                importlib.import_module(KFP_ARTIFACT_ONTOLOGY_MODULE),
                type_name)
            result = artifact_cls()
        except (AttributeError, ImportError, ValueError) as err:
            logging.warning('Failed to instantiate Ontology Artifact:{} \
        instance'.format(type_name))

        if not result:
            # Otherwise generate a generic Artifact object.
            result = Artifact(instance_schema=artifact.type.instance_schema)
        result.runtime_artifact = artifact
        result.metadata = json_format.MessageToDict(artifact.metadata)
        return result

    @classmethod
    def deserialize(cls, data: str) -> Any:
        """Deserializes an Artifact object from JSON dict."""
        artifact = pipeline_spec_pb2.RuntimeArtifact()
        json_format.Parse(data, artifact, ignore_unknown_fields=True)
        return cls.get_from_runtime_artifact(artifact)
